package org.openxma.dsl.generator.ant;

import static org.openxma.dsl.common.DslConstants.OPENXMA_DOM_FILE_EXTENSION;
import static org.openxma.dsl.common.DslConstants.OPENXMA_JSF_POM_FILE_EXTENSION;
import static org.openxma.dsl.common.DslConstants.OPENXMA_POM_FILE_EXTENSION;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Properties;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.types.CommandlineJava;
import org.apache.tools.ant.types.FileSet;
import org.apache.tools.ant.types.Path;
import org.apache.tools.ant.types.Reference;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.mwe.core.WorkflowContext;
import org.eclipse.emf.mwe.core.WorkflowContextDefaultImpl;
import org.eclipse.emf.mwe.core.issues.IssuesImpl;
import org.eclipse.emf.mwe.core.issues.MWEDiagnostic;
import org.eclipse.emf.mwe.core.monitor.NullProgressMonitor;
import org.eclipse.emf.mwe.core.monitor.ProgressMonitor;
import org.eclipse.emf.mwe.core.resources.CachingResourceLoaderImpl;
import org.eclipse.emf.mwe.core.resources.ResourceLoaderFactory;
import org.eclipse.emf.mwe.core.resources.ResourceLoaderImpl;
import org.eclipse.emf.mwe.utils.StandaloneSetup;
import org.eclipse.xtext.ISetup;
import org.eclipse.xtext.mwe.NameBasedFilter;
import org.eclipse.xtext.mwe.SlotEntry;
import org.openxma.dsl.core.model.Model;
import org.openxma.dsl.dom.DomDslStandaloneSetup;
import org.openxma.dsl.generator.DefaultGeneratorConfiguration;
import org.openxma.dsl.generator.DefaultGeneratorFactory;
import org.openxma.dsl.generator.Generator;
import org.openxma.dsl.generator.GeneratorConfiguration;
import org.openxma.dsl.generator.GeneratorConstants;
import org.openxma.dsl.generator.component.DslReader;
import org.openxma.dsl.pom.PomDslStandaloneSetup;
import org.openxma.dsl.pom.PomPackage;

import at.spardat.xma.guidesign.util.ClassLoaderFactory;

import com.google.inject.Injector;

/**
 * Ant task for transforming one or more XMAdsl models to XMA.<br>
 * <h2>Usage:</h2> <h3>Taskdef:</h3>
 * 
 * <pre>
 * &lt;taskdef name         = "dslGenerator"
 *          classname    = "org.openxma.dsl.generator.ant.DslGeneratorTask"
 *          classpathref = "classpath.id" /&gt;
 * </pre>
 * 
 * <h3>Classpath:</h3> The classpath must contain<br>
 * <ul>
 * <li>all standalone XMA dsl jars (e.g. <code>org.openxma.dsl.*.jar</code>)</li>
 * <li>the XMA guidesigner jar (in reality we need the XMA model classes from there) (e.g.
 * <code>at.spardat.xma.guidesigner*.jar</code>)</li>
 * <li>the XMA dsl platform jar (e.g. <code>org.openxma.dsl.platform.*.jar</code>) - attention, this is not a plugin</li>
 * <li>the project output directory, containing the XMAdsl file(s) (xText takes all from classpath)</li>
 * <li>the EMF runtime jars (e.g. <code>org.eclipse.emf*.jar</code>)</li>
 * <li>all the Xtext, Xtend and antlr plugins (e.g.
 * <code>org.eclipse.xtext*.jar:org.eclipse.xtend*.jar:org.antlr.runtime*.jar</code>)</li>
 * <li>Google collect and guice plugins (e.g. <code>com.google.collect*.jar:com.google.guice&#42;/*&#42;/*.jar</code>)</li>
 * <li>Apache cmd line tools plugin (e.g. <code>org.apache.commons.cli*.jar</code>)</li>
 * <li>Apache Ant and Apache Commons plugins (e.g. <code>org.apache&#42;/*&#42;/*.jar:org.apache.commons.*.jar</code>)</li>
 * <li>misc Eclipse core stuff (
 * 
 * <code>org.eclipse.core.jobs*.jar:org.eclipse.core.resources*.jar:org.eclipse.core.runtime*.jar:org.eclipse.osgi*.jar:org.eclipse.osgi*.jar:org.eclipse.equinox.common*.jar:org.eclipse.swt*.jar:com.ibm.icu*.jar</code>
 * )<br>
 * </li>
 * </ul>
 * 
 * <h3>Usage with a file set including all supported model types:</h3>
 * 
 * <pre>
 * &lt;dslGenerator classpathref="classpath.id"
 *         sourceDirectory      = "src/main/java"
 *         genSourceDirectory   = "src/generated/java"
 *         resourceDirectory    = "src/generated/java"
 *         genResourceDirectory    = "src/main/resources"
 *         resourceGenSourceFolder = "src/generated/resources"
 *         templateAdviceFiles     = "templatefiel1.xxy,template2.yyx"
 *         guiDesignProperties     = "src/main/resources/guidesign.properties"
 * &gt;
 *     &lt;classpath&gt;
 *         &lt;path refid = "classpath.id" /&gt;
 *     &lt;/classpath&gt;
 *     &lt;fileset dir="src" casesensitive="yes" includes="&#42;&#42;/&#42;.dml,&#42;&#42;/&#42;.pml,&#42;&#42;/&#42;.xmadsl,&#42;&#42;/&#42;.xma" &gt;
 *         ...
 *     &lt;/fileset&gt;
 * &lt;/xmadsl&gt;
 * </pre>
 * 
 * @author laslovd
 */
public class DslGeneratorTask extends Task {
    static {
        System.setProperty("org.eclipse.emf.ecore.EPackage.Registry.INSTANCE",
                "org.eclipse.emf.ecore.impl.EPackageRegistryImpl");
    }
    private File sourceDirectory;
    private File genSourceDirectory;
    private File testDirectory;
    private File genTestDirectory;
    private File resourceDirectory;
    private File genResourceDirectory;
    private File generatorProperties;
    private String guiDesignProperties;
    private File webAppDirectory;
    private File genWebAppDirectory;
    private boolean useJsfGenerator;

    private final List<FileSet> fileSets = new ArrayList<FileSet>();
    private CommandlineJava commandLine = new CommandlineJava();

    /**
     * The generator properties file.
     * 
     * @parameter generatorProperties
     */
    public void setGeneratorProperties(File generatorProperties) {
        this.generatorProperties = generatorProperties;
    }

    /**
     * The project relative path to the guidesign properties file.
     * 
     * @parameter generatorProperties
     */
    public void setGuiDesignProperties(String guiDesignProperties) {
        this.guiDesignProperties = guiDesignProperties;
    }

    /**
     * Set the classpath to use by reference.
     * 
     * @param reference
     *            a reference to an existing classpath.
     */
    public void setClasspathRef(Reference reference) {
        createClasspath().setRefid(reference);
    }

    /**
     * Set the classpath to be used when running the Dsl generator.
     * 
     * @param s
     *            an Ant Path object containing the classpath.
     */
    public void setClasspath(Path s) {
        createClasspath().append(s);
    }

    /**
     * Add a path to the classpath.
     * 
     * @return created classpath.
     */
    public Path createClasspath() {
        return getCommandLine().createClasspath(getProject()).createPath();
    }

    /**
     * Accessor to the command line.
     * 
     * @return the current command line.
     */
    public CommandlineJava getCommandLine() {
        return commandLine;
    }

    /**
     * For multiple file execution
     * 
     * @param fileset
     */
    public void addFileset(FileSet fileset) {
        fileSets.add(fileset);
    }

    /**
     * the project java source folder (also destination for the generated xma files) <br>
     * optional, default = "src"
     */
    public void setSourceDirectory(File javaSourceFolder) {
        this.sourceDirectory = javaSourceFolder;
    }

    /**
     * the project java source folder for generated artefacts<br>
     */
    public void setGenSourceDirectory(File javaGenSourceFolder) {
        this.genSourceDirectory = javaGenSourceFolder;
    }

    /**
     * the project resource source folder<br>
     */
    public void setResourceDirectory(File resourceSourceFolder) {
        this.resourceDirectory = resourceSourceFolder;
    }

    /**
     * the project resource source folder for generated artefacts<br>
     */
    public void setGenResourceDirectory(File resourceGenSourceFolder) {
        this.genResourceDirectory = resourceGenSourceFolder;
    }

    /**
     * the project test source folder
     */
    public void setTestDirectory(File testDirectory) {
        this.testDirectory = testDirectory;
    }

    /**
     * the project test source folder for generated test artefacts<br>
     */
    public void setGenTestDirectory(File generatedTestDirectory) {
        this.genTestDirectory = generatedTestDirectory;
    }

    public void setUseJsfGenerator(boolean useJsfGenerator) {
        this.useJsfGenerator = useJsfGenerator;
    }

    public void setWebAppDirectory(File webAppDirectory) {
        this.webAppDirectory = webAppDirectory;
    }

    public void setGenWebAppDirectory(File genWebAppDirectory) {
        this.genWebAppDirectory = genWebAppDirectory;
    }

    @Override
    public void execute() throws BuildException {
        checkRequiredParameters();
        ISetup setup = setup();
        IssuesImpl issues = new IssuesImpl();

        DslReader reader = new DslReader();
        reader.setPathTraverser(new DslReader.FileSetTraverser());
        reader.addRegister(setup);
        reader.addRegister(new DomDslStandaloneSetup());
        for (FileSet fileSet : this.fileSets) {
            String[] includedFiles = fileSet.getDirectoryScanner().getIncludedFiles();
            for (String path : includedFiles) {
                reader.addPath(new File(fileSet.getDirectoryScanner().getBasedir(), path).getPath());
            }
        }
        for (String classPathEntry : getCommandLine().getClasspath().list()) {
            reader.addPath(new File(classPathEntry).getPath());
        }
        SlotEntry slotEntry = new SlotEntry();
        slotEntry.setSlot("model");
        slotEntry.setType(Model.class.getSimpleName());
        reader.addLoad(slotEntry);
        NameBasedFilter nameBasedFilter = new NameBasedFilter();
        nameBasedFilter.setRegularExpression("^(?:(?!classes|bin).)*\\.(dml|pml|xmadsl|xma)$");
        reader.setUriFilter(nameBasedFilter);

        WorkflowContext workflowContext = new WorkflowContextDefaultImpl();

        NullProgressMonitor progressMonitor = new NullProgressMonitor();
        reader.invoke(workflowContext, progressMonitor, issues);

        if (!issues.hasErrors()) {
            for (MWEDiagnostic mweDiagnostic : issues.getWarnings()) {
                log(mweDiagnostic.toString(), Project.MSG_WARN);
            }
            for (MWEDiagnostic mweDiagnostic : issues.getInfos()) {
                log(mweDiagnostic.toString(), Project.MSG_INFO);
            }
        }
        @SuppressWarnings("unchecked")
        Collection<Model> models = (Collection<Model>) workflowContext.get("model");

        if (models != null) {
            generate(models, progressMonitor, OPENXMA_DOM_FILE_EXTENSION);
            generate(models, progressMonitor, useJsfGenerator ? OPENXMA_JSF_POM_FILE_EXTENSION
                    : OPENXMA_POM_FILE_EXTENSION);
        }
    }

    private void generate(Collection<Model> models, ProgressMonitor progressMonitor, String fileExtension) {
        DefaultGeneratorConfiguration generatorConfiguration = createGeneratorConfiguration(fileExtension);
        Generator generator = new DefaultGeneratorFactory().createAndInit(generatorConfiguration);
        for (Model model : models) {
            Resource resource = model.eResource();
            generator.generate(resource, progressMonitor);
        }
    }

    private ISetup setup() {
        ClassLoader classloader = createClassloader();
        ResourceLoaderFactory.setCurrentThreadResourceLoader(new CachingResourceLoaderImpl(new ResourceLoaderImpl(
                classloader)));
        Thread.currentThread().setContextClassLoader(classloader);
        ClassLoaderFactory.setProjectClassLoader(classloader);
        StandaloneSetup standaloneSetup = new StandaloneSetup();
        standaloneSetup.setPlatformUri("..");
        standaloneSetup.addRegisterGeneratedEPackage(PomPackage.class.getName());
        final Injector injector = new PomDslStandaloneSetup().createInjectorAndDoEMFRegistration();
        return new ISetup() {
            public Injector createInjectorAndDoEMFRegistration() {
                return injector;
            }
        };
    }

    private ClassLoader createClassloader() {
        List<URL> classLoaderUrls = new ArrayList<URL>();
        for (String classPathEntry : getCommandLine().getClasspath().list()) {
            try {
                classLoaderUrls.add(new File(classPathEntry).toURL());
            } catch (MalformedURLException malformedURLException) {
                log("Exception while create classloader for generator.", malformedURLException, Project.MSG_ERR);
            }
        }
        URL[] urls = classLoaderUrls.toArray(new URL[classLoaderUrls.size()]);
        return new URLClassLoader(urls, getClass().getClassLoader());
    }

    private void checkRequiredParameters() throws BuildException {
        if (sourceDirectory == null) {
            throw new BuildException("Missing attribute 'sourceDirectory'");
        } else if (!sourceDirectory.exists()) {
            throw new BuildException("SourceDirectory '" + sourceDirectory + "' does not exist");
        }
        if (genSourceDirectory == null) {
            genSourceDirectory = sourceDirectory;
        }
        if (resourceDirectory == null) {
            throw new BuildException("Missing attribute 'resourceDirectory'");
        } else if (!resourceDirectory.exists()) {
            throw new BuildException("ResourceDirectory '" + resourceDirectory + "' does not exist");
        }
        if (genResourceDirectory == null) {
            genResourceDirectory = resourceDirectory;
        }
        if (guiDesignProperties == null) {
            log("Missing 'guiDesignProperties' file configuration which is required for core xma generator.");
        }
    }

    protected DefaultGeneratorConfiguration createGeneratorConfiguration(String fileExtension) {
        Properties properties = loadProperties();
        DefaultGeneratorConfiguration generatorConfiguration = new DefaultGeneratorConfiguration(properties);
        generatorConfiguration.setProjectFolder(URI.createURI(getProject().getBaseDir().toURI().toString()));
        generatorConfiguration.setFileExtension(fileExtension);
        generatorConfiguration.setGenResourceFolder(URI.createURI(genResourceDirectory.toURI().toString()));
        generatorConfiguration.setGenSourceFolder(URI.createURI(genSourceDirectory.toURI().toString()));
        generatorConfiguration.setResourceFolder(URI.createURI(resourceDirectory.toURI().toString()));
        generatorConfiguration.setSourceFolder(URI.createURI(sourceDirectory.toURI().toString()));
        if (genTestDirectory != null) {
            generatorConfiguration.setGenTestSourceFolder(URI.createURI(genTestDirectory.toURI().toString()));
        }
        if (testDirectory != null) {
            generatorConfiguration.setTestSourceFolder(URI.createURI(testDirectory.toURI().toString()));
        }
        if (webAppDirectory != null) {
            generatorConfiguration.setWebAppFolder(URI.createURI(webAppDirectory.toURI().toString()));
        }
        if (genWebAppDirectory != null) {
            generatorConfiguration.setWebAppFolder(URI.createURI(genWebAppDirectory.toURI().toString()));
        }
        GeneratorConfiguration.INSTANCE.set(generatorConfiguration);
        return generatorConfiguration;
    }

    protected Properties loadProperties() {
        Properties properties = new Properties();
        if (guiDesignProperties != null) {
            properties.setProperty(GeneratorConstants.GUIDESIGNER_PROPERTIES_FILE, guiDesignProperties);
        }
        if (generatorProperties != null) {
            log("load generator properties from '" + generatorProperties.getPath() + "'", Project.MSG_INFO);
            try {
                properties.load(new FileInputStream(generatorProperties));
            } catch (FileNotFoundException e) {
                throw new BuildException(e.getMessage(), e);
            } catch (IOException e) {
                throw new BuildException(e.getMessage(), e);
            }
        }
        return properties;
    }

}
