package org.openxma.dsl.generator.component;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.mwe.core.WorkflowInterruptedException;
import org.eclipse.emf.mwe.core.issues.Issues;
import org.eclipse.xtext.mwe.PathTraverser;
import org.eclipse.xtext.mwe.Reader;
import org.eclipse.xtext.mwe.Validator;
import org.eclipse.xtext.resource.IResourceServiceProvider;
import org.eclipse.xtext.resource.IResourceServiceProvider.Registry;
import org.eclipse.xtext.validation.CheckMode;
import org.eclipse.xtext.validation.Issue;
import org.openxma.dsl.common.DslConstants;
import org.openxma.dsl.common.logging.LogUtil;

import com.google.common.base.Predicate;
import com.google.common.collect.Lists;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;

public class DslReader extends Reader {

    private Validator validator = new XmaValidator();
    private PathTraverser pathTraverser = new PathTraverser();

    /**
     * Fix URL-Encoded paths which may be contained in classpath if a file name contains e.g. spaces.
     */
    @Override
    public void addPath(String path) {
        String decodedPath = fixPath(path);
        super.addPath(decodedPath);
    }

    protected void populateResourceSet(ResourceSet set, Multimap<String, URI> uris) {
        Collection<URI> values = Sets.newHashSet(uris.values());
        for (URI uri : values) {
            set.getResource(uri, true);
        }
    }

    private String fixPath(String path) {
        String decodedPath;
        try {
            decodedPath = java.net.URLDecoder.decode(path, "utf-8");
            if (!path.equals(decodedPath)) {
                LogUtil.info("URLDecoded path: " + decodedPath);
            }
        } catch (UnsupportedEncodingException e) {
            LogUtil.warn("Unable to URLDecode path: " + path);
            e.printStackTrace();
            decodedPath = path;
        }
        return decodedPath;
    }

    @Override
    protected Validator getValidator() {
        return validator;
    }

    public void setValidator(Validator validator) {
        this.validator = validator;
    }

    protected PathTraverser getPathTraverser() {
        return pathTraverser;
    }

    public void setPathTraverser(PathTraverser pathTraverser) {
        this.pathTraverser = pathTraverser;
    }

    public static class XmaValidator extends Validator {
        @Override
        public void validate(ResourceSet resourceSet, Registry registry, Issues issues) {
            List<Resource> resources = Lists.newArrayList(resourceSet.getResources());
            for (Resource resource : resources) {
                if (resource.getURI().fileExtension().equals(DslConstants.OPENXMA_XMA_FILE_EXTENSION)) {
                    continue;
                }
                try {
                    resource.load(null);
                    IResourceServiceProvider provider = registry.getResourceServiceProvider(resource.getURI());
                    if (provider != null) {
                        List<Issue> result = provider.getResourceValidator().validate(resource, CheckMode.ALL, null);
                        for (Issue issue : result) {
                            switch (issue.getSeverity()) {
                            case ERROR:
                                issues.addError(issue.getMessage(), issue);
                                break;
                            case WARNING:
                                issues.addWarning(issue.getMessage(), issue);
                                break;
                            case INFO:
                                issues.addInfo(issue.getMessage(), issue);
                                break;
                            }
                        }
                    }
                } catch (IOException e) {
                    throw new WorkflowInterruptedException("Couldn't load resource (" + resource.getURI() + ")", e);
                }
            }
            if (isStopOnError() && issues.hasErrors()) {
                String errorMessage = toString(issues);
                throw new WorkflowInterruptedException("Validation problems: \n" + errorMessage);
            }
        }
    }

    public final static class FileSetTraverser extends PathTraverser {
        public Set<URI> findAllResourceUris(String path, Predicate<URI> isValidPredicate) {
            Set<URI> uris = Collections.emptySet();
            File file = new File(path);
            if (file.isFile()) {
                URI uri = URI.createFileURI(file.getAbsolutePath());
                if (isValidPredicate.apply(uri)) {
                    uris = Collections.singleton(uri);
                } else if (file.getName().endsWith(".jar") || file.getName().endsWith(".zip")) {
                    uris = traverseArchive(file, isValidPredicate);
                }
            } else if (file.isDirectory()) {
                uris = traverseDir(file, isValidPredicate);
            }
            return uris;
        }
    }

}
