package org.openxma.dsl.generator.helper;

import java.util.Comparator;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.openxma.dsl.core.model.Model;
import org.openxma.dsl.dom.model.Attribute;
import org.openxma.dsl.dom.model.ComplexType;
import org.openxma.dsl.dom.model.Entity;
import org.openxma.dsl.dom.model.Mapper;
import org.openxma.dsl.dom.model.PropertyMapping;

public class MapperExtension {

    public static Attribute getDerivedAttributeFromMapperRight(Attribute attribute, Mapper mapper) {
        return getDerivedAttribute(attribute, mapper.getRight());
    }

    public static Attribute getDerivedAttributeFromMapperLeft(Attribute attribute, Mapper mapper) {
        return getDerivedAttribute(attribute, mapper.getLeft());
    }

    private static Attribute getDerivedAttribute(Attribute attribute, ComplexType complexType) {
        for (Iterator<Attribute> iter = complexType.getAttributes().iterator(); iter.hasNext();) {
            Attribute next = iter.next();
            if (null == next.getOriginalName()) {
                if (next.getName().equals(attribute.getName()))
                    break;
                continue;
            }

            if (next.getOriginalName().equals(attribute.getName())) {
                ComplexType type = (ComplexType) attribute.eContainer();
                if (next.getOriginalContainerName().equals(type.getName())) {
                    return next;
                }
            }
        }
        return attribute;
    }

    public static String getContainer(Attribute attribute) {
        return attribute.eContainer().toString();
    }

    public static Boolean dataViewIsPropertyMappingSource(ComplexType dataView, Mapper mapper,
            PropertyMapping propertyMapping) {
        return propertyMapping.isBiDirectional() || (propertyMapping.isToLeft() && dataViewIsRight(dataView, mapper))
                || (propertyMapping.isToRight() && dataViewIsLeft(dataView, mapper));
    }

    public static Boolean dataViewIsLeft(ComplexType dataView, Mapper mapper) {
        return mapperInvolvesDataViewAsLeft(mapper, dataView);
    }

    public static Boolean dataViewIsRight(ComplexType dataView, Mapper mapper) {
        return mapperInvolvesDataViewAsRight(mapper, dataView);
    }

    public static Boolean dataViewIsMapperSource(ComplexType dataView, Mapper mapper) {
        boolean isSource = mapper.isBiDirectional()
                || (mapper.isToLeft() && mapperInvolvesDataViewAsRight(mapper, dataView))
                || (mapper.isToRight() && mapperInvolvesDataViewAsLeft(mapper, dataView));

        if (mapper.getPropertyMappings().isEmpty())
            return isSource;
        else {
            for (PropertyMapping propertyMapping : mapper.getPropertyMappings()) {
                if (dataViewIsPropertyMappingSource(dataView, mapper, propertyMapping))
                    return true;
            }
        }

        return false;
    }

    public static String getOppositeElementFullyQualifiedName(ComplexType dataView, Mapper mapper) {
        if (mapperInvolvesDataViewAsLeft(mapper, dataView)) {
            return NameExtension.getFullyQualifiedName(mapper.getRight(), false);
        } else if (mapperInvolvesDataViewAsRight(mapper, dataView)) {
            return NameExtension.getFullyQualifiedName(mapper.getLeft(), false);
        }
        throw new IllegalStateException("dataview not mapped by mapper");
    }

    public static String getOppositeElementName(ComplexType dataView, Mapper mapper) {
        if (mapperInvolvesDataViewAsLeft(mapper, dataView)) {
            return mapper.getRight().getName();
        } else if (mapperInvolvesDataViewAsRight(mapper, dataView)) {
            return mapper.getLeft().getName();
        }
        throw new IllegalStateException("dataview not mapped by mapper");
    }

    public static Set<Mapper> getMappersByDataView(ComplexType dataView) {
        Model model = (Model) dataView.eContainer();
        Set<Mapper> mappers = extractNameSpaceMappers(model.eResource().getResourceSet().getResources().iterator(),
                model.getName());
        return extractRelevantMappersForDataView(mappers, dataView);
    }

    private static Set<Mapper> extractRelevantMappersForDataView(Set<Mapper> mappers, ComplexType dataView) {
        Set<Mapper> result = getSortByMapperNameAscSetInstance();
        for (Mapper next : mappers) {
            if (mapperInvolvesDataView(next, dataView))
                result.add(next);
        }

        return result;
    }

    private static boolean mapperInvolvesDataView(Mapper mapper, ComplexType dataView) {
        return mapperInvolvesDataViewAsLeft(mapper, dataView) || mapperInvolvesDataViewAsRight(mapper, dataView);
    }

    private static boolean mapperInvolvesDataViewAsRight(Mapper mapper, ComplexType dataView) {
        return mapper.getRight().getName().equals(dataView.getName())
                || (mapper.getRight() instanceof Entity && ((Entity) mapper.getRight()).getDefaultDataViewName()
                        .equals(dataView.getName()));
    }

    private static boolean mapperInvolvesDataViewAsLeft(Mapper mapper, ComplexType dataView) {
        return mapper.getLeft().getName().equals(dataView.getName())
                || (mapper.getLeft() instanceof Entity && ((Entity) mapper.getLeft()).getDefaultDataViewName().equals(
                        dataView.getName()));
    }

    private static Set<Mapper> getSortByMapperNameAscSetInstance() {
        return new TreeSet<Mapper>(new Comparator<Mapper>() {
            public int compare(Mapper dv1, Mapper dv2) {
                return dv1.getName().compareTo(dv2.getName());
            }
        });
    }

    private static Set<Mapper> extractNameSpaceMappers(Iterator<Resource> iter, String nameSpace) {
        Set<Mapper> result = getSortByMapperNameAscSetInstance();
        while (iter.hasNext()) {
            Resource resource = iter.next();
            if (resource.getEObject("/") instanceof Model) {
                Model nextModel = (Model) resource.getEObject("/");
                if (nextModel.getName().equals(nameSpace))
                    extractNameSpaceMappers(nextModel, result);
            }
        }
        return result;
    }

    private static Set<Mapper> extractNameSpaceMappers(Model nameSpace, Set<Mapper> result) {
        for (TreeIterator<EObject> eAllContents = nameSpace.eAllContents(); eAllContents.hasNext();) {
            EObject next = eAllContents.next();
            if (next instanceof Mapper)
                result.add((Mapper) next);
        }
        return result;
    }

}
