package org.openxma.dsl.generator.impl;

import static com.google.common.collect.Iterables.filter;
import static com.google.common.collect.Maps.uniqueIndex;
import static org.eclipse.xtext.EcoreUtil2.getAllContentsOfType;
import static org.openxma.dsl.common.extensions.EObjectPropertiesAdapter.hasProperty;
import static org.openxma.dsl.common.extensions.EObjectPropertiesAdapter.setProperty;

import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.xtext.util.SimpleCache;
import org.openxma.dsl.common.Strings;
import org.openxma.dsl.core.model.Constraint;
import org.openxma.dsl.core.model.Model;
import org.openxma.dsl.core.model.XmadslVariable;
import org.openxma.dsl.dom.DomFactory;
import org.openxma.dsl.dom.model.Attribute;
import org.openxma.dsl.dom.model.AttributeProperty;
import org.openxma.dsl.dom.model.AttributeValidationProperty;
import org.openxma.dsl.dom.model.Dao;
import org.openxma.dsl.dom.model.DataView;
import org.openxma.dsl.dom.model.Entity;
import org.openxma.dsl.dom.model.FeatureReference;
import org.openxma.dsl.dom.model.QueryOperation;
import org.openxma.dsl.dom.model.RequiredFlag;

import com.google.common.base.Function;
import com.google.common.base.Predicate;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Lists;

/**
 * Represents a <b>Switch</b> Function for the domain model inheritance hierarchy used to derive default dataViews and
 * to resolve resolve included {@link FeatureReference FeatureReference's} (i.e. convert FeatureReferences to
 * Attributes)
 */
public class DefaultDataViewModelModifier extends AbstractDomModelModifier<Void> implements Function<Entity, DataView> {
    private SimpleCache<Entity, DataView> dataViewCache = new SimpleCache<Entity, DataView>(this);

    public DefaultDataViewModelModifier() {
        setDelegate(new DefaultRepositoryModelModifier());
    }

    @Override
    public Void caseDataView(DataView dataView) {
        resolveDataView(dataView);
        return super.caseDataView(dataView);
    }

    @Override
    public Void caseEntity(Entity entity) {
        apply(entity);
        return super.caseEntity(entity);
    }

    @Override
    public Void caseDao(Dao dao) {
        for (QueryOperation queryOperation : dao.getQueryOperation()) {
            caseQueryOperation(queryOperation);
        }
        return super.caseDao(dao);
    }

    @Override
    public Void caseQueryOperation(QueryOperation queryOperation) {
        if (queryOperation.getType() != null) {
            doSwitch(queryOperation.getType());
        }
        return super.caseQueryOperation(queryOperation);
    }

    public DataView apply(Entity entity) {
        DataView defaultDataView = entity.getDefaultDataView();
        if (defaultDataView == null) {
            defaultDataView = DomFactory.eINSTANCE.createDefaultDataView(entity);
            if (entity.eContainer() instanceof Model) {
                Model model = (Model) entity.eContainer();
                model.getElements().add(defaultDataView);
            }
        }
        return resolveDataView(defaultDataView);
    }

    public DataView resolveDataView(DataView dataView) {
        if (!isAlreadyResolved(dataView) && !dataView.eIsProxy()) {
            markAsResolved(dataView);
            HashMultimap<String, FeatureReference> attributeNameToFeaturesMap = HashMultimap.create();

            for (FeatureReference featureReference : Lists.newArrayList(dataView.getFeatureReferences())) {
                if (featureReference.isAll()) {
                    includeAllFeatureReferences(dataView, featureReference, attributeNameToFeaturesMap);
                } else if (!featureReference.getAttribute().isReference()) {
                    String name = featureReference.getName() != null ? featureReference.getName() : featureReference
                            .getAttribute().getName();
                    attributeNameToFeaturesMap.put(name, featureReference);
                }
            }
            for (FeatureReference featureReference : dataView.getFeatureReferences()) {
                if (null != featureReference.getAttribute()) {
                    Attribute attribute = featureReference.getAttribute();
                    addDataViewAttribute(attributeNameToFeaturesMap, dataView, featureReference,
                            (Attribute) EcoreUtil.copy(attribute));
                }
            }
            Map<String, Attribute> attributeNames = uniqueIndex(dataView.getAttributes(), new Function<Attribute,String>(){
                public String apply(Attribute attribute) {
                    return attribute.getName();
                }
            });
            for (Attribute attribute : dataView.getAttributes()) {
                if (attribute.isReference()) {
                    attribute.setDataType(dataViewCache.get((Entity) attribute.getDataType()));
                } else {
                    filterAttributeConstraints(attributeNames, attribute);
                }
            }
        }
        return dataView;
    }

    protected void filterAttributeConstraints(Map<String, Attribute> attributeNames, Attribute attribute) {
        for (Constraint constraint : getAllContentsOfType(attribute, Constraint.class)) {
            if (constraint.getCondition() != null) {
                Iterable<XmadslVariable> xmadslVariables = filter(getAllContentsOfType(constraint.getCondition(), XmadslVariable.class), new Predicate<XmadslVariable>(){
                    public boolean apply(XmadslVariable xmadslVariable){
                        return xmadslVariable.getReference() instanceof Attribute;
                        
                    }
                });
                for (XmadslVariable xmadslVariable : xmadslVariables) {
                    Attribute constraintAttribute = (Attribute) xmadslVariable.getReference();
                    Attribute dataViewAttribute = attributeNames.get(constraintAttribute.getName());
                    if (dataViewAttribute == null) {
                        AttributeValidationProperty attributeValidationProperty = (AttributeValidationProperty) constraint.eContainer();
                        attributeValidationProperty.getConstraints().remove(constraint);
                        break;
                    } 
                    xmadslVariable.setReference(dataViewAttribute);
                }
            }
        }
    }

    private void markAsResolved(DataView dataView) {
        setProperty(dataView, "resolveIncludedFeatures", Boolean.TRUE);
    }

    private boolean isAlreadyResolved(DataView dataView) {
        return null == dataView || hasProperty(dataView, "resolveIncludedFeatures");
    }

    private void includeAllFeatureReferences(DataView dataView, FeatureReference featureReference,
            HashMultimap<String, FeatureReference> attributeNameToFeaturesMap) {
        List<FeatureReference> resolvedFeatureReferences = Lists.newArrayList();
        for (Attribute attribute : featureReference.getSource().getAllAttributes()) {
            if (!attribute.isReference()) {
                FeatureReference resolvedFeatureReference = DomFactory.eINSTANCE.createFeatureReference();
                resolvedFeatureReference.setSource(featureReference.getSource());
                resolvedFeatureReference.setAttribute(attribute);
                resolvedFeatureReferences.add(resolvedFeatureReference);
                attributeNameToFeaturesMap.put(attribute.getName(), resolvedFeatureReference);
            } else if (!attribute.isMany()) {
                Entity entity = (Entity) attribute.getDataType();
                if (entity.getIdentifier() != null) {
                    FeatureReference resolvedFeatureReference = DomFactory.eINSTANCE.createFeatureReference();
                    resolvedFeatureReference.setName(attribute.getName()
                            + entity.getIdentifier().getName().substring(0, 1).toUpperCase()
                            + entity.getIdentifier().getName().substring(1));
                    resolvedFeatureReference.setSource(entity);
                    setProperty(resolvedFeatureReference, "sourceReference", attribute);
                    resolvedFeatureReference.setAttribute(entity.getIdentifier());
                    resolvedFeatureReferences.add(resolvedFeatureReference);
                    attributeNameToFeaturesMap.put(resolvedFeatureReference.getName(), resolvedFeatureReference);
                } else if (entity.getKey() != null) {
                    // TODO
                }
            }
        }
        dataView.getFeatureReferences().addAll(resolvedFeatureReferences);
    }

    private void addDataViewAttribute(HashMultimap<String, FeatureReference> attributeNameToFeaturesMap,
            DataView dataView, FeatureReference featureReference, Attribute attribute) {
        // remove all attributes except the ones required for validation
        Iterator<AttributeProperty> iterator = attribute.getAttributProperties().iterator();
        while (iterator.hasNext()) {
            AttributeProperty prop = iterator.next();
            if (!(prop instanceof RequiredFlag) && !(prop instanceof AttributeValidationProperty)) {
                iterator.remove();
            }
        }
        attribute.setOriginalName(attribute.getName());
        attribute.setOriginalContainerName(featureReference.getSource().getName());
        attribute.setName(getUniqueAttributeName(attributeNameToFeaturesMap, featureReference, attribute));
        if (attribute.isReference()) {
            Entity entity = (Entity) featureReference.getAttribute().getDataType();
            DataView referenceDataView = featureReference.getView();
            if (referenceDataView == null) {
                referenceDataView = dataViewCache.get(entity);
            }
            featureReference.setView(referenceDataView);
            attribute.setDataType(referenceDataView);
        }
        featureReference.setTarget(attribute);
        // TODO add additional attribute properties specified in data view (featureReference.getAttributProperties())?
        dataView.getAttributes().add(attribute);
    }

    private String getUniqueAttributeName(HashMultimap<String, FeatureReference> attributeNameToFeaturesMap,
            FeatureReference featureReference, Attribute attribute) {
        String uniqueAttributeName = featureReference.getName() != null ? featureReference.getName() : attribute
                .getName();
        if (attributeNameToFeaturesMap.get(uniqueAttributeName).size() > 1) {
            String prefix = Strings.uncapitalize(featureReference.getSource().getName());
            uniqueAttributeName = prefix + attribute.getName().substring(0, 1).toUpperCase()
                    + attribute.getName().substring(1);
        }
        return uniqueAttributeName;
    }

}
