package org.openxma.dsl.generator.impl;

import static org.openxma.dsl.common.extensions.EObjectPropertiesAdapter.setProperty;

import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.xtext.EcoreUtil2;
import org.openxma.dsl.common.Strings;
import org.openxma.dsl.core.CoreFactory;
import org.openxma.dsl.dom.DomFactory;
import org.openxma.dsl.dom.model.Attribute;
import org.openxma.dsl.dom.model.CrudOperationType;
import org.openxma.dsl.dom.model.Dao;
import org.openxma.dsl.dom.model.DaoOperation;
import org.openxma.dsl.dom.model.DataView;
import org.openxma.dsl.dom.model.DelegateOperation;
import org.openxma.dsl.dom.model.Entity;
import org.openxma.dsl.dom.model.FeatureReference;
import org.openxma.dsl.dom.model.Operation;
import org.openxma.dsl.dom.model.Parameter;
import org.openxma.dsl.dom.model.QueryOperation;
import org.openxma.dsl.dom.model.QueryParameter;
import org.openxma.dsl.dom.model.Service;
import org.openxma.dsl.generator.ModelModifier;

import com.google.common.base.Predicate;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;

/**
 * The default {@link ModelModifier Service ModelModifier} mainly used to expand
 * and resolve <b>CRUD</b> operations and to derive required dependencies to
 * other services or repositories. This <code>ModelModifier</code> is also used to
 * enforce interface
 */
public class DefaultServiceModelModifier extends AbstractDomModelModifier<Void> {
	private DefaultDataViewModelModifier dataViewModelModifier = new DefaultDataViewModelModifier();

	public DefaultServiceModelModifier() {
		super();
		setDelegate(dataViewModelModifier);
	}

	@Override
	public Void caseService(Service service) {
		expandCrudOperations(service);
		resolveCrudOperations(service);
		replaceEntitiesWithDataViews(service);
		return super.caseService(service);
	}

	private void expandCrudOperations(Service service) {
		List<DelegateOperation> crudOperations = Lists.newArrayList();
		for (final DelegateOperation delegateOperation : service.getDelegateOperations()) {
			if (CrudOperationType.ALL.equals(delegateOperation.getCrudOperationType())) {
				crudOperations.add(newCrudOperation(delegateOperation, CrudOperationType.CREATE));
				crudOperations.add(newCrudOperation(delegateOperation, CrudOperationType.READ));
				crudOperations.add(newCrudOperation(delegateOperation, CrudOperationType.UPDATE));
				if (!containsDeleteOperation(crudOperations, delegateOperation.getRepository())) {
					crudOperations.add(newCrudOperation(delegateOperation, CrudOperationType.DELETE));
				}
			} else if (!CrudOperationType.NULL.equals(delegateOperation.getCrudOperationType())) {
				if (delegateOperation.getView() == null) {
					delegateOperation.setView(dataViewModelModifier.apply(delegateOperation.getRepository().getEntity()));
				}
				if (delegateOperation.getViewParameter() == null) {
					delegateOperation
							.setViewParameter(dataViewModelModifier.apply(delegateOperation.getRepository().getEntity()));
				}
			}
		}
		if (!crudOperations.isEmpty()) {
			service.getDelegateOperations().addAll(crudOperations);
		}
	}

	private DelegateOperation newCrudOperation(DelegateOperation delegateOperation, CrudOperationType crudOperationType) {
		DelegateOperation crudOperation = DomFactory.eINSTANCE.createDelegateOperation();
		crudOperation.setCrudOperationType(crudOperationType);
		crudOperation.setView(delegateOperation.getView() != null ? delegateOperation.getView() : dataViewModelModifier
				.apply(delegateOperation.getRepository().getEntity()));
		crudOperation.setViewParameter(delegateOperation.getViewParameter() != null ? delegateOperation.getViewParameter()
				: dataViewModelModifier.apply(delegateOperation.getRepository().getEntity()));
		crudOperation.setRepository(delegateOperation.getRepository());
		return crudOperation;
	}
	
	private boolean containsDeleteOperation(List<DelegateOperation> crudOperations, final Dao dao) {
		boolean existingDeleteOperation = Iterables.any(crudOperations, new Predicate<DelegateOperation>() {
			public boolean apply(DelegateOperation input) {
				return CrudOperationType.DELETE.equals(input.getCrudOperationType())
						&& input.getRepository().equals(dao);
			}
		});
		return existingDeleteOperation;
	}

	private void replaceEntitiesWithDataViews(Service service) {
		for (Operation operation : service.getOperations()) {
			if (operation.getType() instanceof Entity) {
				operation.setType(dataViewModelModifier.apply((Entity) operation.getType()));
			}
			EList<Parameter> parameters = operation.getParameters();
			for (Parameter parameter : parameters) {
				if (parameter.getType() instanceof Entity) {
					setProperty(parameter, "entity", parameter.getType());
					parameter.setType(dataViewModelModifier.apply((Entity) parameter.getType()));
				}
			}
		}
	}

	private void resolveCrudOperations(Service service) {
		List<Operation> delegateOperations = Lists.newArrayList();
		for (final DelegateOperation delegateOperation : service.getDelegateOperations()) {
			if (!service.getDependencies().contains(delegateOperation.getRepository())) {
				service.getDependencies().add(delegateOperation.getRepository());
			}
			Entity entity = delegateOperation.getRepository().getEntity();
			if (CrudOperationType.CREATE.equals(delegateOperation.getCrudOperationType())) {
				Operation operation = DomFactory.eINSTANCE.createOperation();
				operation.setDelegate(delegateOperation);
				operation.setType(delegateOperation.getView());
				operation.setName("save");
				Parameter parameter = DomFactory.eINSTANCE.createParameter();
				parameter.setType(delegateOperation.getViewParameter());
				parameter.setName(Strings.uncapitalize(parameter.getType().getName()));
				operation.getParameters().add(parameter);
				delegateOperations.add(operation);
				addServiceDependenciesForDataView(service, entity, (DataView) parameter.getType());
			} else if (CrudOperationType.READ.equals(delegateOperation.getCrudOperationType())) {
				Operation operation = DomFactory.eINSTANCE.createOperation();
				operation.setDelegate(delegateOperation);
				operation.setType(delegateOperation.getView());
				operation.setName("load" + operation.getType().getName());
				addIdOrKeyAsParameter(entity, operation);
				delegateOperations.add(operation);
			} else if (CrudOperationType.UPDATE.equals(delegateOperation.getCrudOperationType())) {
				Operation operation = DomFactory.eINSTANCE.createOperation();
				operation.setDelegate(delegateOperation);
				operation.setName("update");
				Parameter parameter = DomFactory.eINSTANCE.createParameter();
				parameter.setType(delegateOperation.getViewParameter());
				parameter.setName(Strings.uncapitalize(delegateOperation.getViewParameter().getName()));
				operation.getParameters().add(parameter);
				delegateOperations.add(operation);
			} else if (CrudOperationType.DELETE.equals(delegateOperation.getCrudOperationType())) {
				Operation operation = DomFactory.eINSTANCE.createOperation();
				operation.setDelegate(delegateOperation);
				operation.setName("delete" + entity.getName());
				addIdOrKeyAsParameter(entity, operation);
				delegateOperations.add(operation);
			} else if (delegateOperation.getOperation() != null) {
			    Iterable<? extends DaoOperation> daoOperationsWithSameName = getDaoOperationsWithSameName(delegateOperation);
			    for (DaoOperation daoOperation   : daoOperationsWithSameName) {
			        Operation operation = (Operation) copy(daoOperation);
	                operation.setDelegate(delegateOperation);
	                delegateOperations.add(operation);
	                if (delegateOperation.getView() != null) {
	                    operation.setType(delegateOperation.getView());
	                } else if (daoOperation.getType() instanceof Entity) {
	                    delegateOperation.setView(dataViewModelModifier.apply((Entity) operation.getType()));
	                }
	                if (delegateOperation.getName()!=null) {
	                    operation.setName(delegateOperation.getName());
	                }
	                if (delegateOperation.getFilter()!=null) {
	                    Operation filterOperation = (Operation) EcoreUtil.copy(operation);
	                    filterOperation.setDelegate(delegateOperation);
	                    addFilterExpressionParameter(service, filterOperation);
	                    delegateOperations.add(filterOperation);
	                }                    
                }				
			}
		}
		if (!delegateOperations.isEmpty()) {
			service.getOperations().addAll(delegateOperations);
		}
	}

    private Iterable<? extends DaoOperation> getDaoOperationsWithSameName(final DelegateOperation delegateOperation) {
        Dao dao = EcoreUtil2.getContainerOfType(delegateOperation.getOperation(), Dao.class);
        Set<DaoOperation> ops = Sets.newLinkedHashSet();
        ops.addAll(dao.getOperations());
        ops.addAll(dao.getQueryOperation());			    
        Iterable<? extends DaoOperation> daoOperationsWithSameName = Iterables.filter(ops,
                new Predicate<DaoOperation>() {
                    public boolean apply(DaoOperation input) {
                        return input.getName().equals(delegateOperation.getOperation().getName());
                    }
                });
        return daoOperationsWithSameName;
    }

    private void addFilterExpressionParameter(Service service, Operation filterOperation) {
        // remove firstResult and maxResults params since they must be expressed with the following org.openxma.dsl.platform.query.Expression parameter
        Iterator<Parameter> iterator = filterOperation.getParameters().iterator();
        while (iterator.hasNext()) {
            Parameter parameter = iterator.next();
            if ("firstResult".equalsIgnoreCase(parameter.getName())
                    || "maxResults".equalsIgnoreCase(parameter.getName())) {
                iterator.remove();
            }
        }
        Parameter filterParameter = DomFactory.eINSTANCE.createParameter();
        filterParameter.setType(CoreFactory.eINSTANCE.createSimpleType("Expression", "org.openxma.dsl.platform.query.Expression"));
        filterParameter.setName("_filter");
        filterOperation.getParameters().add(filterParameter);
        service.eResource().getContents().add(filterParameter.getType());
    }
	
	private void addServiceDependenciesForDataView(Service service, Entity entity, DataView view) {
		LinkedHashSet<Attribute> processedReferences = Sets.newLinkedHashSet();
		addServiceDependenciesForDataView(service, entity, view, processedReferences);
	}
	
	
    private void addServiceDependenciesForDataView(Service service, Entity referenceOwner) {
        for (Attribute feature : referenceOwner.getAllAttributes()) {
            if (!feature.isReference())
                continue;
            
            Entity entity = (Entity) feature.getDataType();

            if (null != entity.getSuperType() && entity.getSuperType() != referenceOwner)
                addServiceDependenciesForDataView(service, entity.getSuperType());

            Dao dao = entity.getRepository();
            if (null != dao)
                service.getDependencies().add(dao);
        }
    }
	
	private void addServiceDependenciesForDataView(Service service, Entity referenceOwner, DataView view,
			LinkedHashSet<Attribute> processedReferences) {
	        addServiceDependenciesForDataView(service, referenceOwner);
	        
//        for (Attribute feature : referenceOwner.getAttributes()) {
//            if (feature.getDataType() instanceof Entity) {
//                logger.debug("adding depency for entity '" + feature.getName() + "', required: '"
//                        + feature.isRequired() + "'");
//                
//                Dao dao = ((Entity) feature.getDataType()).getRepository();
//                if (null != dao) 
//                    service.getDependencies().add(dao);
//            }
//        }
        
		for (FeatureReference featureReference : view.getFeatureReferences()) {
			if (featureReference.getAttribute() != null && featureReference.getAttribute().isReference()
					&& featureReference.getAttribute().isMany() && processedReferences.add(featureReference.getAttribute())) {
				Entity entity = (Entity) featureReference.getAttribute().getDataType();
				DataView dataView = featureReference.getView() != null ? featureReference.getView()
						: this.dataViewModelModifier.apply(featureReference.getSource());
				if (dataView.includesAllRequiredFeaturesFor(entity)) {
					Dao dao = entity.getRepository();
					if (!service.getDependencies().contains(dao)) {
						service.getDependencies().add(dao);
					}
					addServiceDependenciesForDataView(service, entity, dataView, processedReferences);
				}
			}
		}
	}

	private void addIdOrKeyAsParameter(Entity entity, Operation operation) {
		if (entity.getIdentifier() != null) {
			Parameter parameter = DomFactory.eINSTANCE.createParameter();
			parameter.setType(entity.getIdentifier().getDataType());
			parameter.setName(Strings.uncapitalize(entity.getIdentifier().getName()));
			operation.getParameters().add(parameter);
		} else {
			EList<Attribute> keys = entity.getKey().getResolvedAttributeList();
			for (Attribute attribute : keys) {
				Parameter parameter = DomFactory.eINSTANCE.createParameter();
				parameter.setType(attribute.getDataType());
				parameter.setName(Strings.uncapitalize(attribute.getName()));
				operation.getParameters().add(parameter);
			}
		}
	}

	private Operation copy(DaoOperation daoOperation) {
		if (daoOperation instanceof Operation) {
			return (Operation) EcoreUtil.copy(daoOperation);
		} else if (daoOperation instanceof QueryOperation) {
			Operation operation = DomFactory.eINSTANCE.createOperation();
			operation.setType(daoOperation.getType());
			operation.setMany(daoOperation.isMany());
			operation.setName(daoOperation.getName());
			EList<QueryParameter> parameters = ((QueryOperation) daoOperation).getQueryParameters();
			for (QueryParameter queryParameter : parameters) {
				if (queryParameter instanceof Parameter) {
					operation.getParameters().add((Parameter) EcoreUtil.copy(queryParameter));
				} else if (queryParameter.getAttribute() != null) {
					Parameter parameter = DomFactory.eINSTANCE.createParameter();
					parameter.setType(queryParameter.getAttribute().getDataType());
					parameter.setName(queryParameter.getAttribute().getName());
					parameter.setMany(queryParameter.getAttribute().isMany());
					operation.getParameters().add(parameter);
				}
			}
			return operation;
		}
		throw new IllegalStateException("unknown operation type '" + daoOperation + "'");
	}
}
