package org.openxma.dsl.generator.impl;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EObject;
import org.openxma.dsl.generator.Config;
import org.openxma.dsl.generator.helper.PresentationModelExtension;
import org.openxma.dsl.generator.helper.WidgetExtension;
import org.openxma.dsl.pom.model.GuiElement;
import org.openxma.dsl.pom.model.HeightProperty;
import org.openxma.dsl.pom.model.IElementWithLayoutData;
import org.openxma.dsl.pom.model.LayoutDataProperty;
import org.openxma.dsl.pom.model.Page;
import org.openxma.dsl.pom.model.StylePropertyBackgroundColor;
import org.openxma.dsl.pom.model.StylePropertyFont;
import org.openxma.dsl.pom.model.StylePropertyForegroundColor;
import org.openxma.dsl.pom.model.StyleSpecificationProperty;
import org.openxma.dsl.pom.model.WidthProperty;
import org.openxma.dsl.pom.scoping.PomDslScopeProvider;
import org.openxma.dsl.pom.util.StyleAccess;

import at.spardat.xma.guidesign.XMAPage;
import at.spardat.xma.guidesign.XMAWidget;

public class DefaultWidgetStyling {		
	
	class ColorInfo {
		String name;
		Integer red;
		Integer green;
		Integer blue;

		public String getJavaName() {
			StringBuilder javaName= new StringBuilder("color_");
			if (name != null) {
				javaName.append(name);
			} else {
				javaName.append(red+"_"+green+"_"+blue);
			}
			return javaName.toString();
		}
		
		protected String getRGBParameters() {		
			StringBuilder parameters= new StringBuilder();
			parameters.append(red+", "+green+", "+blue);
			return parameters.toString();
		}	
		
		public String getColorAccessMethod() {
			if (name != null) {
				return "getComposite().getDisplay().getSystemColor(SWT.COLOR_"+name.toUpperCase()+")";
			} else {
				return "getComponent().getColor("+getRGBParameters()+")";
			}
		}
		
	}
	
	class FontInfo {
		Set<String> names = new LinkedHashSet<String>();
		Set<String> styles = new LinkedHashSet<String>();
		Integer size;
		
		public String getJavaName() {
			StringBuilder javaName= new StringBuilder("font");
			for (String name : names) {
				javaName.append(name);
			}
			if (size != null && size>0) javaName.append(size);
			for (String style : styles) {
				javaName.append(style);
			}
			return javaName.toString();
		}
		
		public String getFontParameters(String templateName) {
												
			StringBuilder parameters= new StringBuilder();
			if (names.size() > 0) {
				parameters.append("\""+names.iterator().next()+"\", ");
			} else {
				parameters.append(templateName+".getName(), ");				
			}
			if (size != null && size>0) {
				parameters.append(size+", ");
			} else {
				parameters.append(templateName+".getHeight(), ");
			}
			if (styles.size()>0) {
				boolean first=true;
				for (String style : styles) {
					if (!first) parameters.append("|");
					if (style.equalsIgnoreCase("standard")) {
					    style="NORMAL";
					}
					parameters.append("SWT."+style.toUpperCase());
					first=false;
				}
			} else {
				parameters.append(templateName+".getStyle()");
			}
			return parameters.toString();
		}
	}
	
	
	class LazyStringBuilder {
		private StringBuilder buf;
		private Map<String,FontInfo> fonts;
		
		LazyStringBuilder append(String s) {
			if (buf == null) {
				buf = new StringBuilder();
			}
			buf.append(s);
			return this;
		}
		
		
		void putFont(FontInfo font) {
			if (fonts == null) {
				fonts = new LinkedHashMap<String,FontInfo>();
			}
			fonts.put(font.getJavaName(),font);			
		}
		
		boolean hasValue() {
			return buf != null && buf.length()>0;
		}
		
		StringBuilder getBuffer() {
			return buf;
		}
		
		
		Map<String,FontInfo> getFonts() {
			return fonts;
		}		
	}
	
	class SeperatedBuffers {
		

		private Map<String,ColorInfo> colors;		
		LazyStringBuilder foregroundColor = new LazyStringBuilder();
		LazyStringBuilder backgroundColor = new LazyStringBuilder();
		LazyStringBuilder font = new LazyStringBuilder();
		LazyStringBuilder font_suffix = new LazyStringBuilder();
		
		void putColor(ColorInfo colorInfo) {
			if (colors == null) {
				colors = new LinkedHashMap<String,ColorInfo>();
			}
			colors.put(colorInfo.getJavaName(), colorInfo);
		}
		
		Map<String,ColorInfo> getColors() {
			return colors;
		}		
		
		void appendTo(StringBuilder buf) {
			
			
			if (getColors() != null && getColors().size()>0) {
				buf.append(Config.NL).append("        /* Color variable */").append(Config.NL);
				for (ColorInfo colorInfo : getColors().values()) {
					buf.append("        org.eclipse.swt.graphics.Color "+colorInfo.getJavaName()+" = "+colorInfo.getColorAccessMethod()+";").append(Config.NL);
				}
			}					
			
			if (foregroundColor.hasValue()) {
				buf.append(Config.NL).append("        /* Foreground color */").append(Config.NL);
				buf.append(foregroundColor.getBuffer());
			}		
			if (backgroundColor.hasValue()) {
				buf.append(Config.NL).append("        /* Background color */").append(Config.NL);
				buf.append(backgroundColor.getBuffer());
			}						
			if (font.hasValue()) {
				buf.append(Config.NL).append("        /* Font */").append(Config.NL);
				buf.append("        org.eclipse.swt.graphics.FontData templateFontData = getComposite().getFont().getFontData()[0];").append(Config.NL);
				for (String fontName : font.getFonts().keySet()) {
					FontInfo fontInfo = font.getFonts().get(fontName);					
					buf.append("        org.eclipse.swt.graphics.Font "+fontName+"=getComponent().getFont("+fontInfo.getFontParameters("templateFontData")+");").append(Config.NL);
				}		
				buf.append(font.getBuffer());
				if (font_suffix.hasValue()) {
				    buf.append(font_suffix.getBuffer());
				}
			}									
		}
		
		boolean hasValue() {
		    if (backgroundColor.hasValue()) return true;
		    if (foregroundColor.hasValue()) return true;
		    if (font.hasValue()) return true;
		    return false;
		}
	}
	
	protected void genAdditionalCreateWidgetsCodeClient(DomPageLogicAttachmentStrategy pageLogicAttachmentStrategy, XMAPage xmaPage, StringBuilder buf) {
		
		SeperatedBuffers seperatedBuffers = new SeperatedBuffers();		
		
		// Handle pageComposite
		Page dslPage = DomPageLogicAttachmentStrategy.getDslPage(xmaPage);
		if (dslPage != null) {
			Map<Integer, StyleSpecificationProperty> styleProperties = StyleAccess.getCombinedStyleProperty(dslPage);
			if (styleProperties != null) {
    			Collection<StyleSpecificationProperty> styleSpecificationProperties = styleProperties.values();
    		    if (styleSpecificationProperties != null && styleSpecificationProperties.size()>0) {
    			    applyStyles(seperatedBuffers, xmaPage.getComposite(),styleSpecificationProperties,null);
    		    }
			}
		    if (seperatedBuffers.hasValue()) {
		        buf.append("        final Composite " + xmaPage.getComposite().getNamWidget() + " = getComposite();").append(Config.NL);
		    }
		}
		
	    // Handle all other content
		List<EObject> xmaElements = PomDslScopeProvider.getNsElementsInRange(xmaPage, null);
		for (EObject xmaObj : xmaElements) {
			if (xmaObj instanceof XMAWidget) {
				XMAWidget xmaWidget = (XMAWidget)xmaObj;
				GuiElement guiElement = WidgetExtension.getGuiElementProperty((XMAWidget)xmaObj);
				if (guiElement != null) {
					if (PresentationModelExtension.canModelElementHaveStyle(guiElement)) {
			            Map<Integer, StyleSpecificationProperty> styleProperties = StyleAccess.getCombinedStyleProperty(guiElement);
			            if (styleProperties != null) {
    			            Collection<StyleSpecificationProperty> styleSpecificationProperties = styleProperties.values();					    
    					    //System.out.println("GuiElement= "+guiElement+", style= "+objStyles);
    					    if (styleSpecificationProperties != null && styleSpecificationProperties.size()>0) {
    						    applyStyles(seperatedBuffers, xmaWidget, styleSpecificationProperties, guiElement);
    					    }
			            }
					}				
				}
			}
		}
		
		seperatedBuffers.appendTo(buf);
	}	
	
	protected void applyStyles(SeperatedBuffers buf, XMAWidget xmaWidget, Collection<StyleSpecificationProperty> stylePropList, GuiElement guiElement) {
		for (StyleSpecificationProperty styleProp : stylePropList) {
            if (styleProp instanceof StylePropertyBackgroundColor) {
				StylePropertyBackgroundColor sb = (StylePropertyBackgroundColor)styleProp;
				ColorInfo colorInfo = new ColorInfo();						
				if (sb.getColorID() != null) {
					colorInfo.name = sb.getColorID();
				} else {
					colorInfo.red = sb.getRed();
					colorInfo.green = sb.getGreen();							
					colorInfo.blue = sb.getBlue();														
				}
				buf.putColor(colorInfo);						
				buf.backgroundColor.append("        "+xmaWidget.getNamWidget()+"."+"setBackground("+colorInfo.getJavaName()+");").append(Config.NL);						
			} else if (styleProp instanceof StylePropertyForegroundColor) {
				StylePropertyForegroundColor sb = (StylePropertyForegroundColor)styleProp;
				ColorInfo colorInfo = new ColorInfo();						
				if (sb.getColorID() != null) {
					colorInfo.name = sb.getColorID();					
				} else {
					colorInfo.red = sb.getRed();
					colorInfo.green = sb.getGreen();							
					colorInfo.blue = sb.getBlue();														
				}
				buf.putColor(colorInfo);						
				buf.foregroundColor.append("        "+xmaWidget.getNamWidget()+"."+"setForeground("+colorInfo.getJavaName()+");").append(Config.NL);						
			} else if (styleProp instanceof StylePropertyFont) {
				StylePropertyFont fp = (StylePropertyFont)styleProp;
				FontInfo fi = new FontInfo();
				for (String name : fp.getName()) {
					fi.names.add(name);
				}
				if (fp.getSize()>=0) fi.size=fp.getSize();
				for (String fontStyle : fp.getStyles()) {
					fi.styles.add(fontStyle);
				}
				buf.font.putFont(fi);
				buf.font.append("        "+xmaWidget.getNamWidget()+"."+"setFont("+fi.getJavaName()+");").append(Config.NL);
				xmaWidget.getFormData();
				if ( guiElement instanceof IElementWithLayoutData ) {
				    IElementWithLayoutData guiElementWithLayoutData = (IElementWithLayoutData) guiElement;
                    int widthInChar = -1, heightInChar = -1; 
				    for ( LayoutDataProperty loProp : guiElementWithLayoutData.getLayoutProperties() ) {
				        if ( loProp instanceof WidthProperty ) {
				            WidthProperty wProp = (WidthProperty) loProp;
				            widthInChar = wProp.getWidthInChar();
				        } else if ( loProp instanceof HeightProperty ) {
				            HeightProperty hProp = (HeightProperty) loProp;
				            heightInChar = hProp.getHeightInChar();
				        }
				    }
				    if ( widthInChar > 0 || heightInChar > 0 ) {				        
				        buf.font.append("        adjustSize("+xmaWidget.getNamWidget()+","+widthInChar+","+heightInChar+");").append(Config.NL);;
				        
//				        XMAPage xmaPage = xmaWidget.getPageComposite(xmaWidget).getPage();
//				        String formDataInCreateWidgetsStyleDeclared="formDataInCreateWidgetsStyleDeclared";
//				        buf.font.append("        org.eclipse.swt.graphics.GC gc_"+xmaWidget.getNamWidget()+" = new org.eclipse.swt.graphics.GC("+xmaWidget.getNamWidget()+");").append(Config.NL);
//                        if (WidgetExtension.getGeneratorFlag(xmaPage,formDataInCreateWidgetsStyleDeclared) == false) {
//                            buf.font.append("        FormData ");                         
//                            WidgetExtension.setGeneratorFlag(xmaPage, formDataInCreateWidgetsStyleDeclared, true);
//                        } else {
//                            buf.font.append("        ");  
//                        }
//				        buf.font.append("data = ((FormData) "+xmaWidget.getNamWidget()+".getLayoutData());").append(Config.NL);
//				        if ( widthInChar > 0 )
//				            buf.font.append("        data.width  = getScaler().convertXToCurrent(" + widthInChar + " * gc_"+xmaWidget.getNamWidget()+".getFontMetrics().getAverageCharWidth());").append(Config.NL);
//				        if ( heightInChar > 0 )
//				            buf.font.append("        data.height = getScaler().convertYToCurrent(" + heightInChar + " * gc_"+xmaWidget.getNamWidget()+".getFontMetrics().getHeight());").append(Config.NL);
//				        buf.font.append("        gc_"+xmaWidget.getNamWidget()+".dispose();").append(Config.NL);
				    }
				}
				String layoutCode = "        getComposite().layout();";
				if (buf.font_suffix.getBuffer() == null || !buf.font_suffix.getBuffer().toString().contains(layoutCode)) {
				    buf.font_suffix.append(layoutCode).append(Config.NL);
				}
			}
		}
	}	

}
