package org.openxma.dsl.generator.impl

import javax.inject.Inject
import org.eclipse.emf.ecore.resource.Resource
import org.eclipse.xtext.generator.IFileSystemAccess
import org.eclipse.xtext.generator.IGenerator
import org.openxma.dsl.core.model.Model
import org.openxma.dsl.dom.model.Entity
import org.openxma.dsl.dom.model.Attribute
import org.openxma.dsl.dom.model.AttributeTextProperty

import static org.openxma.dsl.generator.GeneratorConstants.*

import static extension org.openxma.dsl.generator.helper.EntityExtension.*
import static extension org.openxma.dsl.generator.lib.ImportAdapter.*
import org.openxma.dsl.generator.FeatureExtensions

class EntityGenerator implements IGenerator {
	@Inject
    extension Names
	@Inject
    extension Files
	@Inject
    extension Features  
	@Inject
	extension FeatureExtensions

	override doGenerate(Resource resource, IFileSystemAccess fsa) {
		val model = resource.getEObject("/") as Model
		model.elements.filter(typeof(Entity)).forEach(
			[
				fsa.generateFile(qualifiedGenInterfaceName.toJavaFilePath, OUTLET_GENERATED_JAVA, toGenInterface)
				fsa.generateFile(qualifiedInterfaceName.toJavaFilePath, OUTLET_JAVA, toInterface)
				fsa.generateFile(qualifiedGenClassName.toJavaFilePath, OUTLET_GENERATED_JAVA, toGenClass)
				fsa.generateFile(qualifiedClassName.toJavaFilePath, OUTLET_JAVA, toClass)
				if (identifier == null && key != null) {
					fsa.generateFile(key.qualifiedKeyClassName.toJavaFilePath, OUTLET_GENERATED_JAVA, toKeyClass)
				}
				fsa.generateFile(attributePropertiesFileName, OUTLET_GENERATED_RESOURCES, toAttributeProperties)
			]
		)
	}

	def CharSequence toKeyClass(Entity it) {
		'''
		
		package classPackageName;
		
		import java.io.Serializable;
		import org.springframework.util.Assert;
		addImportsHere
		
		/**
		 * Represents a composite identifier for entity <code>name</code>.
		 */
		public class key.keyClassName implements Serializable {
			toSerialVersionUID
			FOR attribute : key.attributes.map(e|e.attribute)
			attribute.toField
			ENDFOR
			FOR attribute : key.attributes.map(e|e.attribute)
			
			attribute.toGetterMethod
			attribute.toSetterMethod
			ENDFOR
			
			public key.keyClassName() {
			}
			
			public key.keyClassName(FOR attribute : key.attributes.map(e|e.attribute) SEPARATOR ','attribute.methodParameterENDFOR) {
				FOR attribute : key.attributes.map(e|e.attribute)
				 Assert.notNull(attribute.name, "parameter 'attribute.name' must not be null");
				ENDFOR
				FOR attribute : key.attributes.map(e|e.attribute)
				 this.attribute.name=attribute.name;
				ENDFOR
			}
			
			@Override
			public int hashCode() {
				final int prime = 31;
				int result = 1;
				FOR attribute : key.attributes.map(e|e.attribute)
				result = prime * result + ((attribute.name == null) ? 0 : attribute.name.hashCode());
				ENDFOR
				return result;
			}
			
			@Override
			public boolean equals(Object obj) {
				if (this == obj) {
					return true;
				}
				if (obj == null) {
					return false;
				}
				if (getClass() != obj.getClass()) {
					return false;
				}
				key.keyClassName other = (key.keyClassName) obj;
				FOR attribute : key.attributes.map(e|e.attribute)
				if (attribute.name == null) {
					if (other.attribute.name != null) {
						return false;
					}
				} else if (!attribute.name.equals(other.attribute.name)) {
					return false;
				}
				ENDFOR
				return true;
			}
		
			@Override
			public String toString() {
				return new StringBuilder("name [")
					FOR attribute : key.attributes.map(e|e.attribute) SEPARATOR '.append(",")'
					.append("attribute.name=").append(attribute.name)
					ENDFOR
					.append("]")
					.toString();
			}
		}
		'''
	}

	def CharSequence toAttributeProperties(Entity it) {
		'''
		FOR attribute : allAttributes
		FOR atp : attribute.attributProperties.filter(typeof(AttributeTextProperty)).filter([labelText != null]) 
			attribute.name.title = atp.toUnicode
			ENDFOR     
			FOR atp : attribute.attributProperties.filter(typeof(AttributeTextProperty)).filter([tooltipText != null]) 
			attribute.name.description = atp.tooltipText
			ENDFOR     
			ENDFOR
			
		'''
	}

	def CharSequence toClass(Entity it) {
		'''
			package classPackageName;
			
			import org.springframework.context.annotation.Scope;
			import org.springframework.stereotype.Component;
			addImportsHere
			
			documentation
			@Component("name.toFirstLower()Prototype")
			@Scope("prototype")
			public class className extends genClassName {
				FOR derivedAttribute : attributes.filter([derived])
				IF !derivedAttribute.readOnly
				/**
				 * @param derivedAttribute.name the derivedAttribute.name to set
				 */
				 public void derivedAttribute.mutatorMethodName(derivedAttribute.dataType.qualifiedName derivedAttribute.name) {
				 }
			    ENDIF
				/**
				 * @return the derivedAttribute.name
				 */
				public derivedAttribute.dataType.qualifiedName derivedAttribute.accessorMethodName() {
					return null;
				}
				ENDFOR
			}
		'''
	}

	def CharSequence toInterface(Entity it) {
		'''
			package packageName;
			
			documentation
			public interface interfaceName extends genInterfaceName {
			}
		'''
	}

	def CharSequence toGenInterface(Entity it) {
		'''
			package packageName;
			
			addImportsHere
			
			documentation
			public interface genInterfaceNameif (superType != null) ''' extends superType.qualifiedName'''  {
				FOR attribute : attributes.filter([!reference])

				attribute.accessorMethod
				attribute.mutatorMethod
				ENDFOR
				FOR attribute : attributes.filter([reference])

				attribute.accessorMethod
				attribute.mutatorMethod
				ENDFOR
				IF  key!=null && (superType==null || superType.key==null) && identifier==null && (superType == null || superType.identifier==null)
				addImport(key.qualifiedKeyClassName) getkey.name();
				ENDIF
			}
		'''
	}

	def CharSequence toGenClass(Entity it) {
		'''
			package classPackageName;
			
			addImportsHere
			
			documentation
			toTypeExtension
			public abstract class genClassNameif (superType != null) ''' extends addImport(superType.qualifiedClassName)''' implements qualifiedName {
				generateAttributes
				generateProperties

				IF key!=null && (superType==null || superType.key==null) && identifier==null && (superType == null || superType.identifier==null)
				public key.keyClassName getkey.name() {
					return new key.keyClassName(FOR attribute : key.attributes.map(e|e.attribute) SEPARATOR ","attribute.nameENDFOR);
				}
				
				ENDIF
				IF hasIdentifierOrKey && (superType == null || !superType.hasIdentifierOrKey)
				generateEqualsHashCode
				
				ENDIF
				generateToString
			}
		'''
	}

	def private generateEqualsHashCode(Entity it) {
		'''
			@Override
			public int hashCode() {
				final int prime = 31;
				int result = 1;
				IF identifier != null
				result = prime * result + ((identifier.accessorMethodName() == null) ? 0 : identifier.accessorMethodName().hashCode());
				ELSEIF key != null
				result = prime * result + ((getkey.name() == null) ? 0 : getkey.name().hashCode());
				ENDIF
				return result;
			}
			
			@Override
			public boolean equals(Object obj) {
				if (this == obj) {
					return true;
				}
				if (obj == null) {
					return false;
				}
				if (!(obj instanceof name)) {
					return false;
				}
				name other = (name) obj;
				IF identifier != null
				if (identifier.accessorMethodName() == null) {
					if (other.identifier.accessorMethodName() != null) {
						return false;
					}
				} else if (!identifier.accessorMethodName().equals(other.identifier.accessorMethodName())) {
					return false;
				}
				ELSEIF key != null
				if (getkey.name() == null) {
					if (((className)other).getkey.name() != null) {
						return false;
					}
				} else if (!getkey.name().equals(((className)other).getkey.name())) {
					return false;
				}
				ENDIF
				return true;
			}
		'''
	}

	def private accessorMethod(Attribute it) {
		'''
		documentationIndentedIF many
		addImport("java.util.Set")<type.dataType.qualifiedName> accessorMethodName();ELSE
		type.dataType.qualifiedName accessorMethodName();ENDIF'''
	}

	def private mutatorMethod(Attribute it) {
		if (many) {
			'''
				boolean hasMethodName();
				void addMethodName(type.dataType.qualifiedName name);
				void removeMethodName(type.dataType.qualifiedName name);
				void removeAllMethodName();
			'''
		} else if (!readOnly) {
			'''void mutatorMethodName(type.dataType.qualifiedName name.toFirstLower);'''
		}
	}

	def private hasIdentifierOrKey(Entity it) {
		null != identifier || null != key
	}
	
	def private attributePropertiesFileName(Entity it) {
		(packageName + "." + interfaceName).toFilePath("properties")
	}

	
}
