package org.openxma.dsl.generator.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.xtext.xbase.lib.StringExtensions;
import org.openxma.dsl.core.model.Constraint;
import org.openxma.dsl.core.model.DataTypeAndTypeParameter;
import org.openxma.dsl.core.model.TrueLiteral;
import org.openxma.dsl.core.model.Type;
import org.openxma.dsl.core.model.ValidatorReference;
import org.openxma.dsl.dom.model.Attribute;
import org.openxma.dsl.dom.model.AttributeProperty;
import org.openxma.dsl.dom.model.AttributeValidationProperty;
import org.openxma.dsl.dom.model.ComplexType;
import org.openxma.dsl.dom.model.RequiredFlag;
import org.openxma.dsl.generator.Config;
import org.openxma.dsl.generator.FeatureExtensions;
import org.openxma.dsl.generator.GeneratorConfiguration;
import org.openxma.dsl.generator.helper.SimpleTypeExtension;
import org.openxma.dsl.generator.helper.ValidatorExtension;

import at.spardat.xma.guidesign.BcdValidator;
import at.spardat.xma.guidesign.CustomValidator;
import at.spardat.xma.guidesign.StringValidator;
import at.spardat.xma.guidesign.Validator;
import at.spardat.xma.guidesign.flex.UserDefinedValidator;

import com.google.inject.Inject;

public class Jsr303FeatureExtension extends FeatureExtensions.NullImpl {

    public static final String ANNOTATION_VALIDFORMAT = "org.openxma.dsl.platform.validation.constraints.ValidFormat";
    public static final String ANNOTATION_DECIMALMAX = "javax.validation.constraints.DecimalMax";
    public static final String ANNOTATION_DECIMALMIN = "javax.validation.constraints.DecimalMin";
    public static final String ANNOTATION_DIGITS = "javax.validation.constraints.Digits";
    public static final String ANNOTATION_SIZE = "javax.validation.constraints.Size";
    public static final String ANNOTATION_NOTNULL = "javax.validation.constraints.NotNull";
    public static final String ANNOTATION_NOTEMPTY_STRING = "org.openxma.dsl.platform.validation.constraints.NotEmpty";

    @Inject
    Names names;
    
    @Inject
    ValidatorGenerator validatorGenerator;
    
    public List<String> groupInterfaces = new ArrayList<String>();
    
    @Override
    public CharSequence toTypeExtension(ComplexType type) {
        if (!GeneratorConfiguration.INSTANCE.isJsr303ValidationEnabled()) {
            return null;
        }
        StringBuilder builder = new StringBuilder();
        builder.append("@").append(names.addImport(type.eContainer(), validatorGenerator.getQualifiedGenValidatorAnnotationClassName(type)));
        return builder;
    }

    @Override
    public CharSequence toFieldExtension(Attribute attribute) {
        if (!GeneratorConfiguration.INSTANCE.isJsr303ValidationEnabled()) {
            return null;
        }
        StringBuilder builder = new StringBuilder();
        Map<String, List<String>> validatorMap = createJsr303Validator(attribute);
        for (String groupInterface : groupInterfaces) {
            builder.append("public interface ").append(groupInterface).append(" { }").append(Config.NL);
        }
        
        for (Entry<String, List<String>> e : validatorMap.entrySet()) {
            String annotationType = e.getKey();
            List<String> parameters = e.getValue();
            boolean isList = parameters.size() > 1;
            if (isList) {
                builder.append("@").append(names.addImport(attribute.eContainer(), annotationType)).append(".List({").append(Config.NL);
            }
            for (Iterator<String> paramIterator = parameters.iterator(); paramIterator.hasNext(); ) {
                String param = paramIterator.next();
                if (isList) {
                    builder.append("\t");
                }
                builder.append("@").append(names.addImport(attribute.eContainer(), annotationType));
                if (param != null) {
                    builder.append("(").append(param).append(")");
                }
                if (isList && paramIterator.hasNext()) {
                    builder.append(",");
                }
                builder.append(Config.NL);
            }
            if (isList) {
                builder.append("})").append(Config.NL);
            }
        }
        
        return builder;
    }
    
    public Map<String, List<String>> createJsr303Validator(Attribute att) {
        groupInterfaces.clear();
        Map<String, List<String>> result = new HashMap<String, List<String>>();
        
        DataTypeAndTypeParameter attType = att.getType();
        
        if (attType.getValidatorReferenceDefinition() != null) {
            if (attType.getValidatorReferenceDefinition().isStringValidator()) {
                StringValidator strValidator = (StringValidator) createXMAValidator(attType);
                addAnnotation(result, createStringValidator(att, strValidator, null));
            } else if (attType.getValidatorReferenceDefinition().isNumberValidator()) {
                BcdValidator bcdValidator = (BcdValidator) createXMAValidator(attType);
                addAnnotation(result, createBcdValidator(att, bcdValidator, null));
            }
            
        }
        List<Constraint> validationConstraints = new ArrayList<Constraint>();
        
        for (AttributeProperty property : att.getAttributProperties()) {
            if (property instanceof RequiredFlag && ((RequiredFlag) property).getExpression() instanceof TrueLiteral) {
//                String groupClass = createRequiredGroup(att);
//                String groups = "groups=" + groupClass;
                if (ValidatorExtension.hasStringValidator(attType)) {
                    addAnnotation(result, ANNOTATION_NOTEMPTY_STRING, null, att);
                } else {
                    addAnnotation(result, ANNOTATION_NOTNULL, null, att);
                }
            } else if (property instanceof AttributeValidationProperty) {
                for (Constraint c: ((AttributeValidationProperty) property).getConstraints()) {
                    validationConstraints.add(c);
                }
            }
        }
        
        int idx = 1;
        for (Constraint c : validationConstraints) {
            addAnnotation(result, createValidatorForConstraint(att, c, idx++));
        }
        
        return result;
    }
    

    private Map<String, List<String>> createStringValidator(Attribute att, StringValidator strValidator, Integer idx) {
        Map<String, List<String>> result = new HashMap<String, List<String>>();
        if (strValidator.getCntMinLength() != null || strValidator.getCntMaxLength() != null) {
            AnnotationPropertyMap propertyMap = new AnnotationPropertyMap();
            propertyMap.put("min", strValidator.getCntMinLength());
            propertyMap.put("max", strValidator.getCntMaxLength());
            if (idx != null) {
                propertyMap.put("groups", createConstraintGroup(att, idx));
            }
            addAnnotation(result, ANNOTATION_SIZE, propertyMap, att);
        }
        if (strValidator.getTxtFormatrange() != null) {
            addAnnotation(result, createFormatValidator(att, strValidator.getTxtFormatrange(), idx));
        }
        return result;
    }

    private Map<String, List<String>> createBcdValidator(Attribute att, BcdValidator bcdValidator, Integer idx) {
        Map<String, List<String>> result = new HashMap<String, List<String>>();
        
        boolean isInteger = SimpleTypeExtension.getTypeForName(att.getType(), "Integer") != null;
        boolean isLong = SimpleTypeExtension.getTypeForName(att.getType(), "Long") != null;
        Integer beforeComma = null;
        if (bcdValidator.getCntBeforeComma() != null) {
            beforeComma = Integer.valueOf(bcdValidator.getCntBeforeComma());
            if ((isInteger && beforeComma == 10) || (isLong && beforeComma == 19)) {
                beforeComma = null;
            }
        }
        Integer afterComma = null;
        if (bcdValidator.getCntAfterComma() != null) {
            if (isInteger || isLong) {
                afterComma = null;
            } else {
                afterComma = Integer.valueOf(bcdValidator.getCntAfterComma());
            }
        }
        if (beforeComma != null || afterComma != null) {
            AnnotationPropertyMap propertyMap = new AnnotationPropertyMap();
            propertyMap.put("integer", beforeComma == null ? "Integer.MAX_VALUE" : beforeComma);
            propertyMap.put("fraction", afterComma == null ? (isInteger || isLong ? 0 : "Integer.MAX_VALUE") : afterComma);
            if (idx != null) {
                propertyMap.put("groups", createConstraintGroup(att, idx));
            }
            addAnnotation(result, ANNOTATION_DIGITS, propertyMap, att);
        }
        
        String qntMinVal = bcdValidator.getQntMinVal();
        if (qntMinVal != null) {
            if (isInteger && String.valueOf(Integer.MIN_VALUE).equals(qntMinVal) || (isLong && String.valueOf(Long.MIN_VALUE).equals(qntMinVal))) {
                qntMinVal = null;
            }
        }
        String qntMaxVal = bcdValidator.getQntMaxVal();
        if (qntMaxVal != null) {
            if (isInteger && String.valueOf(Integer.MAX_VALUE).equals(qntMaxVal) || (isLong && String.valueOf(Long.MAX_VALUE).equals(qntMaxVal))) {
                qntMaxVal = null;
            }
        }
        if (qntMinVal != null || qntMaxVal != null) {
            AnnotationPropertyMap propertyMap = new AnnotationPropertyMap();
            propertyMap.put("min", qntMinVal);
            propertyMap.put("max", qntMaxVal);
            if (idx != null) {
                propertyMap.put("groups", createConstraintGroup(att, idx));
            }
            if (qntMinVal != null) {
                addAnnotation(result, ANNOTATION_DECIMALMIN, propertyMap, att);
            }
            if (qntMaxVal != null) {
                addAnnotation(result, ANNOTATION_DECIMALMAX, propertyMap, att);
            }
        }
        return result;
    }
    
    
    private Map<String, List<String>> createUserDefinedValidator(Attribute att, UserDefinedValidator userValidator, Integer idx) {
        Map<String, List<String>> result = new HashMap<String, List<String>>();
        AnnotationPropertyMap propertyMap = new AnnotationPropertyMap();
        String txtParameterString = userValidator.getTxtParameterString();
        if (txtParameterString != null) {
            propertyMap.put("value", "\"" + txtParameterString + "\"");
        }
        String userValidatorClass = names.addImport(att.eContainer(), userValidator.getNamClass().replaceAll("\\^", "")) + ".class";
        propertyMap.put("validatorClass", userValidatorClass);
        if (idx != null) {
            propertyMap.put("groups", createConstraintGroup(att, idx));
        }
        addAnnotation(result, ANNOTATION_VALIDFORMAT, propertyMap, att);
        return result;
    }


    public String createConstraintGroup(Attribute att, Integer idx) {
        String group = "Is" + StringExtensions.toFirstUpper(att.getName()) + "Constraint" + idx + "Enabled";
        groupInterfaces.add(group);
        String groupClass = group + ".class";
        return groupClass;
    }
    
//    public String createRequiredGroup(Attribute att) {
//        String group = "Is" + StringExtensions.toFirstUpper(att.getName()) + "Required";
//        groupInterfaces.add(group);
//        String groupClass = group + ".class";
//        return groupClass;
//    }
    
    private Map<String, List<String>> createFormatValidator(Attribute att, String formatString, Integer idx) {
        Map<String, List<String>> result = new HashMap<String, List<String>>();
        if (formatString != null) {
            AnnotationPropertyMap propertyMap = new AnnotationPropertyMap();
            propertyMap.put("format", "\"" + formatString + "\"");
            if (idx != null) {
                propertyMap.put("groups", createConstraintGroup(att, idx));
            }
            addAnnotation(result, ANNOTATION_VALIDFORMAT, propertyMap, att);
        }
        return result;
    }
    
    private void addAnnotation(Map<String, List<String>> annotationMap, String annotation, AnnotationPropertyMap propertyMap, Attribute att) {
        List<String> list = annotationMap.get(annotation);
        if (list == null) {
            list = new ArrayList<String>();
            annotationMap.put(annotation, list);
        }
        if (propertyMap == null) {
            propertyMap = new AnnotationPropertyMap();
        }
        propertyMap.put("message", createAnnotationMessageCode(annotation, att));
        list.add(propertyMap.toProperties());
    }

    private String createAnnotationMessageCode(String annotation, Attribute att) {
        StringBuilder code = new StringBuilder();
        code.append("\"{validation.").append(annotation.substring(annotation.lastIndexOf('.') + 1));
        Type attDataType = att.getType().getDataType();
        String typeName = attDataType.getName().toString();
        if (attDataType instanceof org.openxma.dsl.core.model.impl.SimpleTypeImpl) {
            String underlyingDataType = ((org.openxma.dsl.core.model.impl.SimpleTypeImpl) attDataType).getInstanceTypeName();
            if (underlyingDataType != null) {
                underlyingDataType = underlyingDataType.substring(underlyingDataType.lastIndexOf('.') + 1);
                if (!typeName.equals(underlyingDataType)) {
                    code.append(".").append(underlyingDataType);
                }
            }
        }
        code.append(".").append(typeName);
        code.append(".").append(names.getQualifiedName(att.eContainer()));
        code.append(".").append(att.getName());
        code.append("}\"");
        String string = code.toString();
        return string;
    }
    
    private void addAnnotation(Map<String, List<String>> annotationMap, String annotation, String properties) {
        List<String> list = annotationMap.get(annotation);
        if (list == null) {
            list = new ArrayList<String>();
            annotationMap.put(annotation, list);
        }
        list.add(properties);
    }

    private void addAnnotation(Map<String, List<String>> result, Map<String, List<String>> createStringValidator) {
        for (Entry<String, List<String>> entry : createStringValidator.entrySet()) {
            String annotation = entry.getKey();
            for (String properties : entry.getValue()) {
                addAnnotation(result, annotation, properties);
            }
        }
        
    }


    private Map<String, List<String>> createCustomValidator(Attribute att, CustomValidator customValidator, Integer idx) {
        return createFormatValidator(att, customValidator.getTxtFormatingString(), idx);
    }
    
    protected Map<String, List<String>> createValidatorForConstraint(Attribute att, Constraint c, int idx) {
        Map<String, List<String>> result = new HashMap<String, List<String>>();
        
        ValidatorReference validatorReference = c.getValidatorReference();
        if (validatorReference != null) {
            Integer validatorIndex = validatorGenerator.isConstraintConditionTrue(c) ? null : idx;
            if (validatorReference.isStringValidator()) {
                StringValidator strValidator =(StringValidator) ValidatorExtension.createXMAValidator(validatorReference);
                addAnnotation(result, createStringValidator(att, strValidator, validatorIndex));
            } else if (validatorReference.isNumberValidator()) {
                BcdValidator bcdValidator = (BcdValidator) ValidatorExtension.createXMAValidator(validatorReference);
                addAnnotation(result, createBcdValidator(att, bcdValidator, validatorIndex));
            } else if (validatorReference.isCustomValidator()) {
                CustomValidator customValidator = (CustomValidator) ValidatorExtension.createXMAValidator(validatorReference);
                addAnnotation(result, createCustomValidator(att, customValidator, validatorIndex));
            } else if (validatorReference.isUserDefinedValidator()) {
                UserDefinedValidator userValidator = (UserDefinedValidator) ValidatorExtension.createXMAValidator(validatorReference);
                addAnnotation(result, createUserDefinedValidator(att, userValidator, validatorIndex));
                
            }
        }
        return result;
    }

    protected Validator createXMAValidator(DataTypeAndTypeParameter attType) {
        return ValidatorExtension.createXMAValidator(attType.getValidatorReferenceDefinition(), attType.getDefinitionStack());
    }
    
    protected static class AnnotationPropertyMap extends LinkedHashMap<String, Object> {
        private static final long serialVersionUID = 0;

        public String toProperties() {
            if (!this.isEmpty()) {
                StringBuilder builder = new StringBuilder();
                boolean first = true;
                for (java.util.Map.Entry<String, Object> entry : entrySet()) {
                    if (entry.getValue() != null) {
                        if (!first) {
                            builder.append(",");
                        }
                        builder.append(entry.getKey()).append("=").append(entry.getValue());
                        first = false;
                    }
                }
                if (!first) {
                    return builder.toString();
                }
            }
            return null;
        }
    }
    
}
