package org.openxma.dsl.generator.impl

import com.google.common.collect.Sets
import java.util.Set
import javax.inject.Inject
import org.eclipse.emf.ecore.EObject
import org.eclipse.emf.ecore.resource.Resource
import org.eclipse.xtext.generator.IFileSystemAccess
import org.eclipse.xtext.generator.IGenerator
import org.openxma.dsl.core.model.Model
import org.openxma.dsl.dom.model.Attribute
import org.openxma.dsl.dom.model.ComplexType
import org.openxma.dsl.dom.model.CrudOperationType
import org.openxma.dsl.dom.model.Dao
import org.openxma.dsl.dom.model.DataView
import org.openxma.dsl.dom.model.Dependant
import org.openxma.dsl.dom.model.Entity
import org.openxma.dsl.dom.model.FeatureReference
import org.openxma.dsl.dom.model.Operation
import org.openxma.dsl.dom.model.Parameter
import org.openxma.dsl.dom.model.QueryOperation
import org.openxma.dsl.dom.model.QueryParameterValue
import org.openxma.dsl.dom.model.SelectStatement
import org.openxma.dsl.dom.model.Service

import static org.openxma.dsl.generator.GeneratorConstants.*
import static org.openxma.dsl.generator.helper.ServiceExtension.*

import static extension org.openxma.dsl.common.extensions.EObjectPropertiesAdapter.*
import static extension org.openxma.dsl.generator.helper.EntityExtension.*
import static extension org.openxma.dsl.generator.lib.ImportAdapter.*
import static extension org.openxma.dsl.generator.lib.IterableExtensions2.*

class ServiceGenerator implements IGenerator {
	@Inject
	extension Names
	@Inject
	extension Files
	@Inject
	extension Features  

	override doGenerate(Resource resource, IFileSystemAccess fsa) {
		val model = resource.getEObject("/") as Model
		model.elements.filter(typeof(Service)).forEach(
			[
				fsa.generateFile(qualifiedGenInterfaceName.toJavaFilePath, OUTLET_GENERATED_JAVA, toGenInterface)
				fsa.generateFile(qualifiedInterfaceName.toJavaFilePath, OUTLET_JAVA, toInterface)
				fsa.generateFile(qualifiedGenClassName.toJavaFilePath, OUTLET_GENERATED_JAVA, toGenClass)
				fsa.generateFile(qualifiedClassName.toJavaFilePath, OUTLET_JAVA, toClass)
			]
		)
	}

	def toGenInterface(Service it) {
		'''
		package packageName;

		addImportsHere
		documentation
		public interface genInterfaceName {
			FOR operation : operations
			operation.toOperationSignature(null);
			ENDFOR
			FOR operation : getCreateUpdateOperations(operations)
			operation.delegate.view.name saveOrUpdate(operation.delegate.viewParameter.name operation.delegate.viewParameter.toVariableName());
			ENDFOR
		}
		'''
	}

	def toInterface(Service it) {
		'''
		package packageName;

		addImportsHere
		
		documentation
		public interface interfaceName extends genInterfaceName {
		}
		'''
	}

	def toClass(Service it) {
		'''
		package classPackageName;

		import org.springframework.stereotype.Service;
		addImportsHere
		
		documentation
		@Service("name.toFirstLower()")
		public class className extends genClassName {
			FOR operation : operations.filter(e|e.delegate==null)
			operation.toOperationSignature("public") {
				// TODO
				throw new IllegalStateException("not yet implemented");
			}
			ENDFOR
		}
		'''
	}
	
	def toGenClass(Service it) {
		'''
		package classPackageName;

		import org.apache.commons.logging.Log;
		import org.apache.commons.logging.LogFactory;
		import org.openxma.dsl.platform.service.Mapper;
		import org.springframework.beans.factory.annotation.Autowired;
		addImportsHere
		documentation
		public abstract class getGenClassName() implements qualifiedName {
			protected final Log logger = LogFactory.getLog(getClass());
			@Autowired(required=false)
			protected Mapper mapper;
			FOR dependency : dependencies
			protected dependency.name dependency.qualifiedName.toFirstLower();
			ENDFOR
			FOR dependency : dependencies
			
			@Autowired
			public void setdependency.name(dependency.qualifiedName dependency.toVariableName()) {
				this.dependency.toVariableName() = dependency.toVariableName();
			}
			ENDFOR
			FOR crudOperation : crudOperations
			
			crudOperation.toCrudOperationImpl
			ENDFOR
			FOR op : getCreateUpdateOperations(operations)
			IF op.delegate.repository.entity.version != null || op.delegate.repository.entity.identifier != null
			
			op.toCrudSaveOrUpdateImplementation
			ENDIF
			
			ENDFOR
			FOR queryOperation : operations.filter(e|e.delegate?.operation != null)
			queryOperation.toQueryOperationImplementation
			
			ENDFOR
		}
		'''
	}
	
	def private toCrudOperationImpl(Operation it) {
		val crudOperationType = delegate?.crudOperationType   
		val repository =  delegate?.repository
		val entity =  repository?.entity
		val viewParameter = delegate?.viewParameter
		switch crudOperationType {
			
			case CrudOperationType::CREATE : {
				val toManyReferences = viewParameter.featureReferences.filter(e|e.source==entity && e.attribute?.reference && e.attribute?.many && e.view?.includesAllRequiredFeaturesFor(e.attribute?.dataType as Entity))
				val optionalReferences = entity?.allNonRequiredReferenceNonManyAttributesRecursive
				'''
				@addImport("org.springframework.transaction.annotation.Transactional")
				toOperationSignature("public") {
					addImport("org.springframework.util.Assert").notNull(viewParameter.toVariableName(), "parameter 'viewParameter.toVariableName()' must not be null");
					entity.qualifiedName entity.toVariableName() = this.mapper.createAndMapOne(viewParameter.toVariableName(), entity.qualifiedName.class,"saveentity.name");
					FOR requiredReference : entity.requiredReferences
					requiredReference.dataType.qualifiedName requiredReference.toVariableName() = this.mapper.createAndMapOne(viewParameter.toVariableName(), requiredReference.dataType.name.class,"requiredReference.name");
					requiredReference.toVariableName() = (requiredReference.dataType as Entity).repository.toVariableName().load(requiredReference.toVariableName().requiredReference.getIdOrKeyAccessorMethodName()());
					entity.toVariableName().requiredReference.getMutatorMethodName()(requiredReference.toVariableName());
					ENDFOR
					FOR optionalReference : optionalReferences
					val target = viewParameter.featureReferences.findFirst(e|e.attribute == optionalReference)
					IF target != null
					if (null != viewParameter.toVariableName().target.getAccessorMethodName()()) {
						optionalReference.dataType.qualifiedName optionaloptionalReference.name.toFirstUpper = this.mapper.createAndMapOne(viewParameter.toVariableName().target.getAccessorMethodName()(), optionalReference.dataType.name.class,"optionalReference.name");
						optionaloptionalReference.name.toFirstUpper = (optionalReference.dataType as Entity).repository.toVariableName().load(optionaloptionalReference.name.toFirstUpper.optionalReference.getIdOrKeyAccessorMethodName()());
						entity.toVariableName().optionalReference.getMutatorMethodName()(optionaloptionalReference.name.toFirstUpper);
					}
					ENDIF
					ENDFOR
					FOR includedFeature : toManyReferences
					addincludedFeature.attribute.name.toFirstUpper(entity.toVariableName(),viewParameter.toVariableName());
					ENDFOR
					this.repository.toVariableName().saveOrUpdate(entity.toVariableName());
					return this.mapper.mapOne(entity.toVariableName(),new type.name());
				}
				val Set<Attribute> processedAttributes = Sets::newHashSet
				FOR includedFeature : toManyReferences
				includedFeature.toAddReferenceMethod(viewParameter, processedAttributes)
				ENDFOR
				'''
			}
			
			case CrudOperationType::READ : {
				'''
				@addImport("org.springframework.transaction.annotation.Transactional")(readOnly = true)
				toOperationSignature("public") {
					IF entity?.identifier != null
						addImport("org.springframework.util.Assert").notNull(entity.identifier.name, "parameter 'entity.identifier.name' must not be null");
						return this.mapper.createAndMapOne(this.repository.toVariableName().load(entity.identifier.name),type.name.class);
					ELSEIF entity?.key != null
						FOR attribute : entity.key.attributes.map(e|e.attribute).<Attribute>toList
						addImport("org.springframework.util.Assert").notNull(attribute.name, "parameter 'attribute.name' must not be null");
						ENDFOR
						return this.mapper.createAndMapOne(this.repository.toVariableName().load(new addImport(entity.key.qualifiedKeyClassName)(FOR field : entity.key.attributes.map(e|e.attribute) SEPARATOR ","field.nameENDFOR)),type.name.class);
					ENDIF
				}
				'''
			}
			
			case CrudOperationType::UPDATE : {
				'''
				@addImport("org.springframework.transaction.annotation.Transactional")
				toOperationSignature("public") {
					addImport("org.springframework.util.Assert").notNull(viewParameter.toVariableName(), "parameter 'viewParameter.toVariableName()' must not be null");
					IF entity?.identifier!=null
					entity.qualifiedName mappedentity.name = this.mapper.createAndMapOne(viewParameter.toVariableName(), entity.name.class);
					entity.name entity.toVariableName() = this.repository.toVariableName().load(mappedentity.name.entity.identifier.getAccessorMethodName()());
					IF entity.version!=null && viewParameter.includesVersionFor(entity) 
					if (!entity.toVariableName().entity.version.getAccessorMethodName()().equals(mappedentity.name.entity.version.getAccessorMethodName()())) {
						throw new addImport("org.springframework.dao.OptimisticLockingFailureException")("Entity 'entity.name' was updated or deleted by another transaction.");
					}
					ENDIF
					ELSE
					entity.qualifiedName entity.toVariableName() = this.repository.toVariableName().repository.naturalKey.getMethodName()(FOR field : entity.key.attributes.map(e|e.attribute) SEPARATOR ","viewParameter.toVariableName().field.getAccessorMethodName()()ENDFOR);
					ENDIF
					this.mapper.mapOne(viewParameter.toVariableName(), entity.toVariableName(),"updateentity.name");
				}
				'''
			}
			
			case CrudOperationType::DELETE : {
				'''
				@addImport("org.springframework.transaction.annotation.Transactional")
				toOperationSignature("public") {
					IF entity?.identifier!=null
					addImport("org.springframework.util.Assert").notNull(entity.identifier.name, "parameter 'entity.identifier.name' must not be null");
					this.repository.toVariableName().delete(entity.identifier.name);
					ELSEIF entity?.key!=null
					FOR field : entity.key.attributes.map(e|e.attribute)
					addImport("org.springframework.util.Assert").notNull(field.name, "parameter 'field.name' must not be null");
					ENDFOR
					this.repository.toVariableName().delete(new addImport(entity.key.qualifiedKeyClassName)(FOR field : entity.key.attributes.map(e|e.attribute) SEPARATOR ","field.nameENDFOR));
					ENDIF
				}
				'''
			}
			
		}
	}
	
	def private toCrudSaveOrUpdateImplementation(Operation it) {
		val repository = delegate.repository
		val view = delegate.view
		val viewParameter = delegate.viewParameter
		'''
		@addImport("org.springframework.transaction.annotation.Transactional")
		public view.name saveOrUpdate(viewParameter.name viewParameter.toVariableName()) {
			addImport("org.springframework.util.Assert").notNull(viewParameter.toVariableName(), "parameter 'viewParameter.toVariableName()' must not be null");
			IF repository.entity.version != null
			if ( viewParameter.toVariableName().repository.entity.version.getAccessorMethodName()() == null ) {
			ELSE
			if ( viewParameter.toVariableName().repository.entity.identifier.getAccessorMethodName()() == null ) {
			ENDIF
				return this.save(viewParameter.toVariableName());
			}
			else {
				this.update(viewParameter.toVariableName());
				IF view == viewParameter
				return viewParameter.toVariableName();
				ELSE
				IF repository.entity.identifier!=null
				repository.entity.qualifiedName mappedrepository.entity.name = this.mapper.createAndMapOne(viewParameter.toVariableName(), repository.entity.name.class);
				repository.entity.name repository.entity.toVariableName() = this.repository.toVariableName().load(mappedrepository.entity.name.repository.entity.identifier.getAccessorMethodName()());
				ELSE
				repository.entity.qualifiedName repository.entity.toVariableName() = this.repository.toVariableName().repository.naturalKey.getMethodName()(FOR field : repository.entity.key.attributes.map(e|e.attribute) SEPARATOR ","viewParameter.toVariableName().field.getAccessorMethodName()()ENDFOR);
				ENDIF
				this.mapper.mapOne(viewParameter.toVariableName(), repository.entity.toVariableName(),"updaterepository.entity.name");
				return this.mapper.createAndMapOne(repository.entity.toVariableName(),view.name.class);
				ENDIF
			}
		}
		
		'''
	}
	
	def private toQueryOperationImplementation(Operation it) {
		'''
		@addImport("org.springframework.transaction.annotation.Transactional")IF (delegate!=null && delegate.operation instanceof QueryOperation && (delegate.operation as QueryOperation).statement instanceof SelectStatement)(readOnly=true)ENDIF
		toOperationSignature("public") {
			IF delegate.filter!=null && delegate.operation instanceof QueryOperation && !parameters.empty && parameters.last().name.equalsIgnoreCase("_filter")
			addImport("org.openxma.dsl.platform.query.QueryObject") queryObject = QueryObject.createNamedQuery(getQualifiedQueryConstant(delegate.operation as QueryOperation), _filter);
			FOR parameter : (delegate.operation as QueryOperation).queryParameters.reject(e|e instanceof Parameter && ((e as Parameter).name.equals("firstResult") || (e as Parameter).name.equals("maxResults")))
			val queryParameterValue = delegate.operation.eAllContents.filter(typeof(QueryParameterValue)).findFirst([value|value.parameter==parameter || value.parameter==parameter.attribute])
			IF parameter.type instanceof ComplexType && queryParameterValue?.attribute != null
			FOR attribute : delegate.operation.eAllContents.filter(typeof(QueryParameterValue)).filter(e|e.parameter==parameter).map(e|e.attribute).<Attribute>toSet().sortBy(e|e.name)
			queryObject.withArgs("parameter.getName()_attribute.name", parameter.getName().attribute.getAccessorMethodName()());
			ENDFOR
			ELSEIF queryParameterValue != null
			 queryObject.withArgs("parameter.getName()", parameter.getName());
			ENDIF
			ENDFOR
			delegate.operation.toType() _result =  this.delegate.repository.toVariableName().query(queryObject);
			IF delegate.view!=null
			return this.mapper.createAndMapMany(_result,delegate.view.qualifiedName.class);
			ELSE
			return _result;
			ENDIF
			ELSE
			IF delegate.view!=null
			return this.mapper.IF manycreateAndMapManyELSEcreateAndMapOneENDIF(this.delegate.repository.toVariableName().delegate.operation.name(FOR parameter : parameters SEPARATOR ","IF parameter.hasProperty("entity")mapper.createAndMapOne(parameter.toVariableName(),parameter.entityType.qualifiedName.class)ELSEparameter.toVariableName()ENDIFENDFOR),delegate.view.qualifiedName.class);
			ELSE
			IF type != nullreturn ENDIFthis.delegate.repository.toVariableName().delegate.operation.name(FOR parameter : parameters SEPARATOR ","IF parameter.hasProperty("entity")mapper.createAndMapIF parameter.manyManyELSEOneENDIF(parameter.toVariableName(),parameter.entityType.qualifiedName.class)ELSEparameter.toVariableName()ENDIFENDFOR);
			ENDIF
			ENDIF
		}
		'''
	}
	
	def EObject getEntityType(Parameter it) {
		getProperty("entity") as EObject
	}
	
	def private getIdOrKeyAccessorMethodName(Attribute it) {
		if (dataType instanceof Entity) {
			val entity = dataType as Entity
			if (entity.identifier != null) {
				return entity.identifier.accessorMethodName
			} else if (entity.key != null) {
				return entity.key.accessorMethodName  
			}
		}
		return accessorMethodName
	}
	
	def private Iterable<Operation> getCrudOperations(Service it) {
		return operations.filter(e|e.delegate?.crudOperationType!=null && e.delegate.crudOperationType!=CrudOperationType::NULL && e.delegate.crudOperationType!=CrudOperationType::ALL)	   
	}
	
	def private String toAddReferenceMethod(FeatureReference it, DataView dataView, Set<Attribute> processedReferences) {
		if (!processedReferences.add(attribute)) {
			return "";
		}
		val String addReferenceMethod = '''
		
		private void addattribute.name.toFirstUpper(source.qualifiedName parentsource.name,dataView.qualifiedName dataView.toVariableName()) {
			if (dataView.toVariableName().attribute.getAccessorMethodName()() != null) {
				for (view.qualifiedName _view.toVariableName() : dataView.toVariableName().attribute.getAccessorMethodName()()) {
					attribute.dataType.qualifiedName attribute.dataType.toVariableName() = this.mapper.createAndMapOne(_view.toVariableName(), attribute.dataType.qualifiedName.class,"saveattribute.dataType.name");
					FOR requiredReference : (attribute.dataType as Entity).requiredReferences.filter(e|e.opposite!=attribute && attribute.opposite!=e)
					val requiredEntity = (requiredReference.dataType as  Entity)
					requiredReference.dataType.qualifiedName requiredReference.toVariableName() = this.mapper.createAndMapOne(_view.toVariableName(), requiredReference.dataType.qualifiedName.class,"requiredReference.name");
					if (requiredReference.toVariableName().requiredEntity.identifier.getAccessorMethodName()() != null) {
						requiredReference.toVariableName() = requiredEntity.repository.toVariableName().load(requiredReference.toVariableName().requiredEntity.identifier.getAccessorMethodName()());
						attribute.dataType.toVariableName().requiredReference.getMutatorMethodName()(requiredReference.toVariableName());
					}
					ENDFOR
					FOR nestedIncludedFeature : view.featureReferences.filter(e|e.attribute?.reference && e.attribute?.many && !processedReferences.contains(e.attribute) && e.view.includesAllRequiredFeaturesFor(e.attribute.dataType as Entity))
					addnestedIncludedFeature.attribute.name.toFirstUpper(attribute.dataType.toVariableName(),_view.toVariableName());
					ENDFOR
					parentsource.name.attribute.getAddMethodName()(attribute.dataType.toVariableName());
				}
			}
		}
		'''
		val StringBuilder result = new StringBuilder(addReferenceMethod)
		for (includedFeature : view.featureReferences.filter(e|e.view != null  && e.attribute?.many && e.view.includesAllRequiredFeaturesFor(e.attribute.dataType as Entity))) {
			result.append(includedFeature.toAddReferenceMethod(view,processedReferences)) 
		}
		result.toString
	}
	
	def private getName(Dependant it) {
		switch it {
			Dao: it.name
			Service: it.name
		}
	}

}
