package org.openxma.dsl.generator.impl

import at.spardat.xma.guidesign.BcdValidator
import at.spardat.xma.guidesign.CustomValidator
import at.spardat.xma.guidesign.StringValidator
import at.spardat.xma.guidesign.flex.UserDefinedValidator
import javax.inject.Inject
import org.eclipse.emf.ecore.resource.Resource
import org.eclipse.xtext.generator.IFileSystemAccess
import org.eclipse.xtext.generator.IGenerator
import org.eclipse.xtext.util.SimpleAttributeResolver
import org.openxma.dsl.core.model.Model
import org.openxma.dsl.core.model.ReferenceableByXmadslVariable
import org.openxma.dsl.core.model.StatusFlag
import org.openxma.dsl.dom.model.AttributeValidationProperty
import org.openxma.dsl.dom.model.Entity
import org.openxma.dsl.dom.model.DataView
import org.openxma.dsl.dom.model.Property
import org.openxma.dsl.dom.model.RequiredFlag
import org.openxma.dsl.generator.lib.Index

import static org.openxma.dsl.generator.GeneratorConstants.*

import static extension org.openxma.dsl.generator.helper.EntityExtension.*
import static extension org.openxma.dsl.generator.helper.SimpleTypeExtension.*
import static extension org.openxma.dsl.generator.helper.ValidatorExtension.*
import static extension org.openxma.dsl.generator.lib.ImportAdapter.*
import org.openxma.dsl.core.model.ModelElement
import org.openxma.dsl.dom.model.ComplexType
import org.openxma.dsl.core.model.Constraint

class ValidatorGenerator implements IGenerator {
	@Inject
	extension Types
	@Inject
	extension Names
	@Inject
	extension Files
	@Inject
	extension Features
	@Inject
	extension ExpressionCompiler

	override doGenerate(Resource resource, IFileSystemAccess fsa) {
		val model = resource.getEObject("/") as Model
		model.elements.filter(typeof(Entity)).forEach(
			[
				fsa.generateFile(qualifiedGenValidatorClassName.toJavaFilePath, OUTLET_GENERATED_JAVA, toGenClass)
				fsa.generateFile(qualifiedValidatorClassName.toJavaFilePath, OUTLET_JAVA, toClass)
				if (isJsr303ValidationEnabled) {
					fsa.generateFile(qualifiedGenValidatorAnnotationClassName.toJavaFilePath, OUTLET_GENERATED_JAVA, toGenAnnotationClass)
				}
			]
		)
		if (isJsr303ValidationEnabled) {
			model.elements.filter(typeof(DataView)).forEach(
				[
					fsa.generateFile(qualifiedGenValidatorClassName.toJavaFilePath, OUTLET_GENERATED_JAVA, toViewGenClass)
					fsa.generateFile(qualifiedGenValidatorAnnotationClassName.toJavaFilePath, OUTLET_GENERATED_JAVA, toGenAnnotationClass)
					fsa.generateFile(qualifiedValidatorClassName.toJavaFilePath, OUTLET_JAVA, toClass)
				]
			)
		}
	}

	def toGenAnnotationClass(ComplexType it) {
		'''
		package validatorPackageName;
		
		import static java.lang.annotation.ElementType.TYPE;
		import static java.lang.annotation.RetentionPolicy.RUNTIME;
		
		import java.lang.annotation.Documented;
		import java.lang.annotation.Retention;
		import java.lang.annotation.Target;
		
		import javax.validation.Constraint;
		import javax.validation.Payload;
		
		@Target({ TYPE })
		@Retention(RUNTIME)
		@Documented
		@Constraint(validatedBy = {validatorClassName.class})
		public @interface genValidatorAnnotationClassName {
			String message() default "{getQualifiedGenValidatorAnnotationClassName.message}";
		
			Class<?>[] groups() default { };
		
			Class<? extends Payload>[] payload() default {};
		
		}
		'''
	}
	
	def toGenClass(Entity it) {
		val applicationContext = requiredApplicationContext
		'''
		package validatorPackageName;
		
		addImportsHere
		IF !isJsr303ValidationEnabledimport static org.openxma.dsl.platform.validation.ValidatorsHelper.*;ENDIF
		
		
		@SuppressWarnings("boxing")
		public abstract class getGenValidatorClassName()IF isJsr303ValidationEnabled extends addImport("org.openxma.dsl.platform.validation.Jsr303Validators")<addImport(getQualifiedGenValidatorAnnotationClassName()), addImport(getQualifiedName())>ELSEIF superType!=null extends addImport(superType.getValidatorPackageName() + "." + superType.getValidatorClassName())ELSE extends addImport("org.openxma.dsl.platform.validation.Validators")ENDIF {
			IF applicationContext!=null
			private applicationContext.qualifiedName context;

			public applicationContext.qualifiedName getContext() {
				return context;
			}

			public void setContext(applicationContext.qualifiedName context) {
				this.context = context;
			}
			ENDIF
			
			IF isJsr303ValidationEnabledcreateJsr303ValidationELSEcreateSpringValidationENDIF
			createConstraintConditions
		}
		'''
	}

	def toClass(ComplexType it) {
		'''
		package getValidatorPackageName();

		addImportsHere
		
		public class getValidatorClassName() extends getGenValidatorClassName() {
		IF !isJsr303ValidationEnabled
			@Override
			public void validate(Object object, addImport("org.springframework.validation.Errors") errors) {
			    super.validate(object, errors);
			    // validate additional constraints
			}
		ENDIF
		}
		'''
	}


	def toViewGenClass(DataView it) {
		'''
		package validatorPackageName;
		
		addImportsHere
		
		@SuppressWarnings("boxing")
		public abstract class getGenValidatorClassName()IF isJsr303ValidationEnabled extends addImport("org.openxma.dsl.platform.validation.Jsr303Validators")<addImport(getQualifiedGenValidatorAnnotationClassName()), addImport(getQualifiedName())>ENDIF {
			
			createJsr303Validation
			createConstraintConditions
		}
		'''
	}

	def private createConstraintConditions(ComplexType it) {
		'''
			FOR function : eAllContents.toList.filter(typeof(ReferenceableByXmadslVariable)).sortBy(e|SimpleAttributeResolver::NAME_RESOLVER.apply(e))
			function.toFunction(it)
			ENDFOR
			FOR attribute : attributes.filter(e|!e.attributProperties.filter(typeof(AttributeValidationProperty)).map(a|a.constraints).isEmpty)
			val index = new Index
			val attributProperties = attribute.attributProperties.filter(typeof(AttributeValidationProperty))
			val constraints = attributProperties.map(a|a.constraints).flatten
			FOR constraint : constraints
			val compiled = compileConstraintCondition(constraint)
			IF !isJsr303ValidationEnabled || !isCompiledConditionTrue(compiled)
			
			public boolean isattribute.name.toFirstUpper()Constraintindex.counter1Enabled(name toVariableName()) {
				return compiled;
			}
			ENDIF
			index.increment
			ENDFOR
			ENDFOR
		'''
	}
	
	def private compileConstraintCondition(Constraint constraint) {
		if (constraint.condition == null) {
			return "Boolean.TRUE";
		} else {
			return constraint.condition.compile;
		}
	}
	
	def public isConstraintConditionTrue(Constraint constraint) {
		return isCompiledConditionTrue(compileConstraintCondition(constraint));
	}
	
	def private static isCompiledConditionTrue(CharSequence conditionString) {
		return "Boolean.TRUE".equals(conditionString);
	}

	def private createSpringValidation(Entity it) {
		'''
			@SuppressWarnings("unchecked")
			public boolean supports(Class clazz) {
				return addImport(getQualifiedInterfaceName()).class.isAssignableFrom(clazz);
			}

			public addImport("org.springframework.validation.Errors") validate(name toVariableName()) {
				Errors errors = createErrors(toVariableName());
				IF superType!=null
				super.validate(toVariableName(),errors);
			    ENDIF
				validate(toVariableName(),errors);
				return errors;
			}

			public void validate(Object object, Errors errors) {
				name toVariableName() = (name) object;
				FOR attribute : attributes.filter(e|!e.attributProperties.filter(typeof(RequiredFlag)).isEmpty)
				IF attribute.required
				addImport("org.springframework.validation.ValidationUtils").rejectIfEmptyIF attribute.type.hasStringValidator()OrWhitespaceENDIF(errors, "attribute.toVariableName()", "field.required");
				ENDIF
				ENDFOR
				FOR attribute : attributes
				IF attribute.type.hasStringValidator()
				val validator = attribute.type.validatorReferenceDefinition.createXMAValidator(attribute.type.definitionStack) as StringValidator
				IF validator.cntMinLength!=null
				rejectIfMinLength(errors,"attribute.name",toVariableName().attribute.getAccessorMethodName()(),validator.cntMinLength);
				ENDIF
				IF validator.cntMaxLength!=null
				rejectIfMaxLength(errors,"attribute.name",toVariableName().attribute.getAccessorMethodName()(),validator.cntMaxLength);
				ENDIF
				ELSEIF attribute.type.hasNumberValidator()
				val validator = attribute.type.validatorReferenceDefinition.createXMAValidator(attribute.type.definitionStack) as BcdValidator
				IF validator.cntBeforeComma!=null && (attribute.type.getTypeForName("Integer")==null || validator.cntBeforeComma.asInteger!=10) && (attribute.type.getTypeForName("Long")==null || validator.cntBeforeComma.asInteger!=19)
				rejectIfMaxIntegerDigits(errors,"attribute.name",toVariableName().attribute.getAccessorMethodName()(),validator.cntBeforeComma);
				ENDIF
				IF validator.cntAfterComma!=null && attribute.type.getTypeForName("Integer")==null && attribute.type.getTypeForName("Long")==null
				rejectIfMaxFractionDigits(errors,"attribute.name",toVariableName().attribute.getAccessorMethodName()(),validator.cntAfterComma);
				ENDIF
				ENDIF
				ENDFOR
				FOR attribute : attributes.filter(e|!e.attributProperties.filter(typeof(AttributeValidationProperty)).map(a|a.constraints).isEmpty)
				val iter = new Index
				FOR constraint : attribute.attributProperties.filter(typeof(AttributeValidationProperty)).map(a|a.constraints).flatten
				IF constraint.validatorReference.isStringValidator()
				val validator = constraint.validatorReference.createXMAValidator() as StringValidator
				if (isattribute.name.toFirstUpper()Constraintiter.counter1Enabled(toVariableName())) {
					IF validator.cntMinLength!=null
					rejectIfMinLength(errors,"attribute.name",toVariableName().attribute.getAccessorMethodName()(),validator.cntMinLength);
					ENDIF
					IF validator.cntMaxLength!=null
					rejectIfMaxLength(errors,"attribute.name",toVariableName().attribute.getAccessorMethodName()(),validator.cntMaxLength);
					ENDIF
					IF validator.txtFormatrange!=null
					rejectIfInvalidFormat(errors,"attribute.name",toVariableName().attribute.getAccessorMethodName()(),"validator.txtFormatrange");
					ENDIF
				}
				ELSEIF constraint.validatorReference.isNumberValidator()
				val validator = constraint.validatorReference.createXMAValidator() as BcdValidator
				if (isattribute.name.toFirstUpper()Constraintiter.counter1Enabled(toVariableName())) {
					IF validator.qntMinVal!=null
					rejectIfMinValue(errors,"attribute.name",toVariableName().attribute.getAccessorMethodName()(),validator.qntMinVal);
					ENDIF
					IF validator.qntMaxVal!=null
					rejectIfMaxValue(errors,"attribute.name",toVariableName().attribute.getAccessorMethodName()(),validator.qntMaxVal);
					ENDIF
					IF validator.cntBeforeComma!=null && (attribute.type.getTypeForName("Integer")==null || validator.cntBeforeComma.asInteger()!=10) && (attribute.type.getTypeForName("Long")==null || validator.cntBeforeComma.asInteger()!=19)
					rejectIfMaxIntegerDigits(errors,"attribute.name",toVariableName().attribute.getAccessorMethodName()(),validator.cntBeforeComma);
					ENDIF
					IF validator.cntAfterComma!=null && attribute.type.getTypeForName("Integer")==null && attribute.type.getTypeForName("Long")==null
					rejectIfMaxFractionDigits(errors,"attribute.name",toVariableName().attribute.getAccessorMethodName()(),validator.cntAfterComma);
					ENDIF
				}
				ELSEIF constraint.validatorReference.isCustomValidator()
				val validator = constraint.validatorReference.createXMAValidator() as CustomValidator
				if (isattribute.name.toFirstUpper()Constraintiter.counter1Enabled(toVariableName())) {
					IF validator.txtFormatingString!=null
					rejectIfInvalidFormat(errors,"attribute.name",toVariableName().attribute.getAccessorMethodName()(),"validator.txtFormatingString");
					ENDIF
				}
				ELSEIF constraint.validatorReference.isUserDefinedValidator()
				val validator = constraint.validatorReference.createXMAValidator() as UserDefinedValidator
				if (isattribute.name.toFirstUpper()Constraintiter.counter1Enabled(toVariableName())) {
					IF validator.namClass!=null
					rejectIfInvalidFormat(errors,"attribute.name",toVariableName().attribute.getAccessorMethodName()(),new validator.namClass.replaceAll("\\^", "")("validator.txtParameterString"));
					ENDIF
				}
				ENDIF
				iter.increment
				ENDFOR
				ENDFOR
			}
		'''
	}

	def private createJsr303Validation(ComplexType it) {
		val groups = createJsr303ValidationGroups(it);
		if (groups.length() > 0) {
		'''
			protected addImport("java.util.Set")<Class<?>> createGroups(addImport(getQualifiedName()) toVariableName()) {
				Set<Class<?>> groups = new addImport("java.util.HashSet")<Class<?>>();
				groups
				return groups;
			}
		'''
		}
	}

	def private createJsr303ValidationGroups(ComplexType it) {
		'''
				FOR attribute : attributes.filter(e|!e.attributProperties.filter(typeof(AttributeValidationProperty)).map(a|a.constraints).isEmpty)
				val iter = new Index
				FOR constraint : attribute.attributProperties.filter(typeof(AttributeValidationProperty)).map(a|a.constraints).flatten
				IF constraint.validatorReference.isStringValidator() || constraint.validatorReference.isNumberValidator()
					|| constraint.validatorReference.isCustomValidator() || constraint.validatorReference.isUserDefinedValidator()
				IF !isConstraintConditionTrue(constraint)
				if (isattribute.name.toFirstUpper()Constraintiter.counter1Enabled(toVariableName())) {
					groups.add(addImport(getQualifiedGenClassName + ".Is" + attribute.name.toFirstUpper() + "Constraint" + iter.counter1 + "Enabled").class);
				}
				ENDIF
				ENDIF
				iter.increment
				ENDFOR
				ENDFOR
		'''
	}


	def dispatch toFunction(ReferenceableByXmadslVariable it, ModelElement element) {
		''''''
	}
	
	def dispatch toFunction(StatusFlag it, ModelElement element) {
		'''
		
		public boolean name.toFirstUpper()(element.qualifiedName element.toVariableName) {
			return expression.compile;
		}
		'''
	}

	def dispatch toFunction(Property it, ModelElement element) {
		'''
		
		public boolean name.toFirstUpper()(element.qualifiedName element.toVariableName) {
			return false;
		}
		'''
	}
	
	def private getQualifiedValidatorClassName(ModelElement it) {
		validatorPackageName + "." + validatorClassName
	}

	def private getQualifiedGenValidatorClassName(ModelElement it) {
		validatorPackageName + "." + genValidatorClassName
	}

	def getQualifiedGenValidatorAnnotationClassName(ModelElement it) {
		validatorPackageName + "." + genValidatorAnnotationClassName
	}

	def private getValidatorClassName(ModelElement it) {
		name + "Validator"
	}

	def private getGenValidatorClassName(ModelElement it) {
		name + "GenValidator"
	}

	def private getGenValidatorAnnotationClassName(ModelElement it) {
		"Valid" + name
	}

	def private getValidatorPackageName(ModelElement it) {
		model.name + ".validation"
	}

	def private asInteger(String value) {
		if (value != null) {
			return Integer::valueOf(value)    
		}
		return 0;
	}

}

