package org.openxma.dsl.generator.impl

import com.google.common.collect.Lists
import java.io.File
import java.util.List
import javax.inject.Inject
import javax.inject.Named
import org.eclipse.core.resources.ResourcesPlugin
import org.eclipse.core.runtime.Path
import org.eclipse.emf.common.notify.Notifier
import org.eclipse.emf.ecore.resource.Resource
import org.eclipse.emf.ecore.resource.ResourceSet
import org.eclipse.emf.mwe.core.monitor.ProgressMonitor
import org.eclipse.xtext.generator.IFileSystemAccess
import org.eclipse.xtext.generator.IFileSystemAccessExtension2
import org.eclipse.xtext.generator.IGenerator
import org.openxma.dsl.core.model.Model
import org.openxma.dsl.generator.Generator
import org.openxma.dsl.generator.GeneratorConfiguration
import org.openxma.dsl.generator.GeneratorConstants
import org.openxma.dsl.generator.IFileSystemAccessProvider
import org.openxma.dsl.generator.component.HbmFileMergePostProcessor
import org.openxma.dsl.generator.lib.ImportAdapter

import static org.apache.commons.logging.LogFactory.*

import static extension org.openxma.dsl.common.extensions.EObjectPropertiesAdapter.*

class XtendDomGenerator implements Generator, IGenerator {

	val logger = getLog(getClass().getName());
	@Property GeneratorConfiguration generatorConfiguration
	@Inject @Named("valueObjectGenerator") IGenerator valueObjectGenerator
	@Inject @Named("repositoryGenerator") IGenerator repositoryGenerator
	@Inject @Named("hbmXmlFileGenerator") IGenerator hbmXmlFileGenerator
	@Inject @Named("entityGenerator") IGenerator entityGenerator
	@Inject @Named("dataTypeGenerator") IGenerator dataTypeGenerator
	@Inject @Named("serviceGenerator") IGenerator serviceGenerator
	@Inject @Named("dataViewGenerator") IGenerator dataViewGenerator
	@Inject @Named("validatorGenerator") IGenerator validatorGenerator
	@Inject @Named("sessionContextGenerator") IGenerator sessionContextGenerator
	@Inject IFileSystemAccessProvider fileSystemAccessProvider;
	ProgressMonitor monitor;

	override init(GeneratorConfiguration generatorConfiguration) {
		this.generatorConfiguration = generatorConfiguration
	}

	override generate(Resource resource, ProgressMonitor monitor) {
		this.monitor = monitor;
        try {
            monitor.beginTask("on " + resource.getURI().lastSegment(), 10)
            val fileSystemAccess = fileSystemAccessProvider.get(generatorConfiguration)
            doGenerate(resource, fileSystemAccess)
        } finally {
            monitor.done
        }
	}

	override doGenerate(Resource resource, IFileSystemAccess fileSystemAccess) {
        if (resource.URI.archive || monitor.canceled) {
            return
        }
        resource.resourceSet.modify
        val fsa = new IFileSystemAccessDelegate(fileSystemAccess, resource.resourceSet)
        subTask("- write ValueObjects", 1)
        valueObjectGenerator.doGenerate(resource, fsa)
        subTask("- write Entities", 1)
        entityGenerator.doGenerate(resource, fsa)
        subTask("- write Repositories", 1)
        repositoryGenerator.doGenerate(resource, fsa)
        subTask("- write ValueObjects", 1)
        hbmXmlFileGenerator.doGenerate(resource, fsa)
        subTask("- write Validators", 1)
        validatorGenerator.doGenerate(resource, fsa)
        subTask("- write DataViews", 1)
        dataViewGenerator.doGenerate(resource, fsa)
        subTask("- write Services", 1)
        serviceGenerator.doGenerate(resource, fsa)
        subTask("- write Context", 1)
        sessionContextGenerator.doGenerate(resource, fsa)
        subTask("- write Types", 1)
        dataTypeGenerator.doGenerate(resource, fsa)
	}

	def modify(ResourceSet resourceSet) {
		val List<Resource> resources = Lists::newArrayList(resourceSet.resources);
		for (Resource resource : resources) {
			if (resource.getEObject("/") instanceof Model && !resource.hasProperty(GeneratorConstants::MODIFIED_MODEL)) {
                subTask(" - pre-process model " + resource.getURI().lastSegment(),0)
				resource.setProperty(GeneratorConstants::MODIFIED_MODEL, Boolean::TRUE);
				logger.info("Pre-process model '" + resource.getURI().lastSegment() + "'");
				generatorConfiguration.modelModifier.modifyModel(resource.getEObject("/") as Model)
			}
		}
	}

	def subTask(String name, int worked) {
		monitor.worked(worked)
		monitor.subTask(name)
	}

}

class IFileSystemAccessDelegate implements IFileSystemAccess {
	val logger = getLog(typeof(XtendDomGenerator).name);
	val IFileSystemAccess delegate
	val Notifier notifier
	val hbmFileMergePostProcessor = new HbmFileMergePostProcessor

	new(IFileSystemAccess delegate, Notifier notifier) {
		this.delegate = delegate
		this.notifier = notifier
	}

	override deleteFile(String fileName) {
		delegate.deleteFile(fileName)
	}

	override generateFile(String fileName, CharSequence contents) {
		generateFile(fileName, DEFAULT_OUTPUT, contents)
	}

	override generateFile(String fileName, String outputConfigurationName, CharSequence contents) {
		logger.info("Write " + fileName)
		delegate.generateFile(fileName, outputConfigurationName, ImportAdapter::resolve(notifier, contents.toString))
		val uri = (delegate as IFileSystemAccessExtension2).getURI(fileName, outputConfigurationName)
		if (uri.platform) {
			val root = ResourcesPlugin::getWorkspace().getRoot();
			val folder = root.getFolder(new Path(uri.toPlatformString(true)));
			hbmFileMergePostProcessor.afterClose(new File(folder.getLocationURI()))
		} else if (uri.file) {
			hbmFileMergePostProcessor.afterClose(new File(uri.toFileString))
		}
	}

}
